# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import pandas as pd


def _check_column_uniqueness(column: "pd.core.series.Series") -> bool:
    r"""
    Check if all elements are equal in the column.

    Arg:
        column (pandas.core.series.Series): Column to check
    """
    first_ele = column[0]
    for ele in column:
        if first_ele != ele:
            return True
    return False


def _get_dynamic_columns(df: pd.DataFrame) -> pd.DataFrame:
    r"""
    Given a ``pd.DataFrame``, return a DataFrame where columns have more than 1 unique value.

    Args:
        df (pd.DataFrame): DataFrame to examine
    """
    hashable_columns = []
    unhashable_columns = []
    for col in df.columns:
        try:
            hash(df[col][0])
            hashable_columns.append(col)
        except TypeError:
            # If a column is not hashable, check if there exists value differ from the
            if _check_column_uniqueness(df[col]):
                unhashable_columns.append(col)

    unique_counts = df[hashable_columns].nunique()
    columns_with_multiple_unique_values = unique_counts[unique_counts > 1].index
    dfs_to_concat = [df[columns_with_multiple_unique_values], df[unhashable_columns]]
    if (
        "prompt" in df
        and "prompt" not in df[columns_with_multiple_unique_values]
        and "prompt" not in df[unhashable_columns]
    ):
        dfs_to_concat.append(df["prompt"])
    elif (
        "messages" in df
        and "messages" not in df[columns_with_multiple_unique_values]
        and "messages" not in df[unhashable_columns]
    ):
        dfs_to_concat.append(df["messages"])
    return pd.concat(dfs_to_concat, axis=1)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


# Sentry collects crash reports and performance numbers
# It is possible to turn off data collection using an environment variable named "SENTRY_OPT_OUT"
import sentry_sdk

import os
import platform
import uuid
import hashlib
from .version import __version__


SENTRY_DSN = "https://43fbb5a3a556ca0a879f5a08ce805d87@o4505656408211456.ingest.sentry.io/4505656412667904"

# Get a random token based on the machine uuid
token = hashlib.sha256(str(uuid.getnode()).encode()).hexdigest()


def find_certifi_path():
    try:
        import certifi

        return os.path.join(os.path.dirname(certifi.__file__), "cacert.pem")
    except Exception:
        pass
    return None


def filter_info(event, _hint):
    # Remove personal info
    try:
        event["modules"] = None
        event["extra"] = None
        event["server_name"] = None
    except Exception:
        pass
    return event


def init_sentry():
    if "SENTRY_OPT_OUT" not in os.environ:
        if platform.system() == "Darwin":
            # Fix CA certificate issue on latest MAC models
            path = find_certifi_path()
            if path is not None:
                if "SSL_CERT_FILE" not in os.environ:
                    os.environ["SSL_CERT_FILE"] = path
                if "REQUESTS_CA_BUNDLE" not in os.environ:
                    os.environ["REQUESTS_CA_BUNDLE"] = path

        sentry_sdk.init(
            dsn=SENTRY_DSN,
            release=__version__,
            traces_sample_rate=0.01,
            include_local_variables=False,
            send_default_pii=False,
            attach_stacktrace=False,
            before_send=filter_info,
            include_source_context=False,
            # the rate at which transaction and performance data is sampled for profiling purposes
            profiles_sample_rate=0.0,
        )
        try:
            filename = os.path.join(os.environ.get("HOME", "/tmp"), ".token")
            if platform.system() == "Windows":
                filename = os.path.join(os.environ.get("USERPROFILE", "c:\\"), ".token")
            with open(filename, "w") as f:
                f.write(token)
        except Exception:
            pass

    sentry_sdk.capture_message("Initializing prompttools", "info")



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Dict, List, Tuple
import itertools
from time import perf_counter
import logging

try:
    from mysql.connector.connection_cext import CMySQLConnection
except ImportError:
    CMySQLConnection = None

from prompttools.mock.mock import mock_mindsdb_completion_fn

from .experiment import Experiment
from .error import PromptExperimentException


class MindsDBExperiment(Experiment):
    r"""
    An experiment class for MindsDB.
    This accepts combinations of MindsDB inputs to form SQL queries, returning a list of responses.

    Args:
        db_connector (CMySQLConnection): Connector MindsDB
        kwargs (dict): keyword arguments for the model
    """

    def __init__(
        self,
        db_connector: "CMySQLConnection",
        **kwargs: Dict[str, object],
    ):
        self.cursor = db_connector.cursor()
        self.completion_fn = self.mindsdb_completion_fn
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_mindsdb_completion_fn

        self.call_params = dict(prompt=kwargs["prompt"])
        self.model_params = dict({k: kwargs[k] for k in kwargs if k != "prompt"})

        self.all_args = self.model_params | self.call_params
        super().__init__()

    def prepare(self) -> None:
        r"""
        Creates argument combinations by taking the cartesian product of all inputs.
        """
        self.model_argument_combos = [
            dict(zip(self.model_params, val)) for val in itertools.product(*self.model_params.values())
        ]
        self.call_argument_combos = [
            dict(zip(self.call_params, val)) for val in itertools.product(*self.call_params.values())
        ]

    def mindsdb_completion_fn(
        self,
        **params: Dict[str, Any],
    ) -> List[Any]:
        r"""
        MindsDB helper function to make request.
        """
        prompt = params["prompt"]

        self.cursor.execute(prompt)
        return [x for x in self.cursor]

    def run(
        self,
        runs: int = 1,
    ) -> None:
        r"""
        Create tuples of input and output for every possible combination of arguments.
        For each combination, it will execute `runs` times, default to 1.
        # TODO This can be done with an async queue
        """
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        results = []
        latencies = []
        for model_combo in self.model_argument_combos:
            for call_combo in self.call_argument_combos:
                call_combo["prompt"] = call_combo["prompt"].format(
                    table=model_combo["table"],
                    author_username=model_combo["author_username"],
                    text=model_combo["text"],
                )
                for _ in range(runs):
                    call_combo["client"] = self.cursor
                    start = perf_counter()
                    res = self.completion_fn(**call_combo)
                    latencies.append(perf_counter() - start)
                    results.append(res)
                    self.argument_combos.append(model_combo | call_combo)
        if len(results) == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException
        self._construct_result_dfs(self.argument_combos, results, latencies, extract_response_equal_full_result=True)

    @staticmethod
    def _extract_responses(output: List[Dict[str, object]]) -> Tuple[str]:
        return output[0]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, List, Optional
from functools import wraps
import logging

from .threshold_type import ThresholdType
from .error.failure import PromptTestSetupException
from .runner.runner import run_prompttest

TESTS_TO_RUN = []


def prompttest(
    metric_name: str,
    eval_fn: Callable,
    prompts: List[str],
    threshold: float = 1.0,
    threshold_type: ThresholdType = ThresholdType.MINIMUM,
    expected: Optional[List[str]] = None,
):
    r"""
    Creates a decorator for prompt tests, which can annotate evaluation functions.
    This enables developers to create a prompt test suite from their evaluations.
    """

    def prompttest_decorator(completion_fn: Callable):
        @wraps(completion_fn)
        def runs_test():
            results = [completion_fn(prompt) for prompt in prompts]
            return run_prompttest(
                metric_name,
                eval_fn,
                threshold,
                threshold_type,
                prompts,
                results,
                expected=expected,
            )

        TESTS_TO_RUN.append(runs_test)
        return runs_test

    return prompttest_decorator


def main():
    logging.getLogger().setLevel(logging.WARNING)
    print("Running " + str(len(TESTS_TO_RUN)) + " test(s)")
    failures = int(sum([test() for test in TESTS_TO_RUN]))
    if failures == 0:
        print("All " + str(len(TESTS_TO_RUN)) + " test(s) passed!")
        exit(0)
    else:
        print("Tests failed: " + str(failures))
        exit(1)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

r"""
These mock functions exist for testing and demo purposes.
"""
import json

try:
    import cv2
except ImportError:
    cv2 = None


class DotDict(dict):
    r"""
    Have dot access to dictionary attributes
    """
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __getattr__(self, key):
        try:
            val = self.get(key)
            if val is None:  # Doesn't support `None` as value
                raise KeyError
            else:
                return val
        except KeyError:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")


def mock_openai_chat_completion_fn(**kwargs):
    return DotDict(
        {
            "choices": [
                DotDict(
                    {
                        "finish_reason": "stop",
                        "index": 0,
                        "message": DotDict(
                            {
                                "content": "George Washington",
                                "role": "assistant",
                            }
                        ),
                    }
                )
            ],
            "created": 1687839008,
            "id": "",
            "model": "gpt-3.5-turbo-0301",
            "object": "chat.completion",
            "usage": DotDict({"completion_tokens": 18, "prompt_tokens": 57, "total_tokens": 75}),
        }
    )


def mock_openai_chat_function_completion_fn(**kwargs):
    return DotDict(
        {
            "choices": [
                DotDict(
                    {
                        "finish_reason": "stop",
                        "index": 0,
                        "message": DotDict(
                            {
                                "role": "assistant",
                                "content": None,
                                "function_call": DotDict(
                                    {
                                        "name": "get_current_weather",
                                        "arguments": '{\n  "location": "Toronto, Canada",\n  "format": "celsius"\n}',
                                    }
                                ),
                            }
                        ),
                    }
                )
            ],
            "created": 1687839008,
            "id": "",
            "model": "gpt-3.5-turbo-0301",
            "object": "chat.completion",
            "usage": DotDict({"completion_tokens": 18, "prompt_tokens": 57, "total_tokens": 75}),
        }
    )


def mock_openai_completion_fn(**kwargs):
    return DotDict(
        {
            "id": "",
            "object": "text_completion",
            "created": 1589478378,
            "model": "text-davinci-003",
            "choices": [
                DotDict(
                    {
                        "text": json.dumps({"text": "George Washington"}),
                        "index": 0,
                        "logprobs": None,
                        "finish_reason": "length",
                    }
                )
            ],
            "usage": DotDict({"prompt_tokens": 5, "completion_tokens": 7, "total_tokens": 12}),
        }
    )


def mock_hf_completion_fn(**kwargs):
    return [{"generated_text": "George Washington"}]


def mock_chromadb_fn(**kwargs):
    return {
        "ids": [["id1"]],
        "embeddings": None,
        "documents": [["George Washington lived in modern day Philadelphia"]],
        "metadatas": [[{"source": "my_source"}]],
        "distances": [[0.5932742953300476]],
    }


class _mock_Anthropic_Completion_Object(object):
    def __init__(self, completion: str, model: str, stop_reason: str):
        self.completion = completion
        self.model = model
        self.stop_reason = stop_reason


def mock_anthropic_completion_fn(**kwargs):
    return _mock_Anthropic_Completion_Object(
        completion='{"name": "John", "age": 30, "city": "New York", "interests": ["reading","hiking","coding"]}',
        model="claude-2.0",
        stop_reason="stop_sequence",
    )


class _mock_PaLM_Completion_Object(object):
    def __init__(self, candidates: list[dict], result: str, filters: list = [], safety_feedback: list = []):
        self.candidates = candidates
        self.result = result
        self.filters = filters
        self.safety_feedback = safety_feedback


def mock_palm_completion_fn(**kwargs):
    return _mock_PaLM_Completion_Object(
        candidates=[{"output": "How are you today?", "safety_ratings": []}], result="How are you today?"
    )


def mock_mindsdb_completion_fn(**kwargs):
    return [
        (
            "The first president of the United States was George Washington. However, "
            "if you're referring to a different country, please specify so I can provide the correct information.",
        )
    ]


def mock_lc_completion_fn(**kwargs):
    return "The first president of the United States was George Washington."


def mock_stable_diffusion(**kwargs):
    if cv2 is None:
        raise ModuleNotFoundError(
            "Package `cv2` is required to be installed to use this experiment."
            "Please use `pip install opencv-python` to install the package"
        )
    return cv2.imread("/mock_data/images/Just_a_fruit_basket.png")


def mock_replicate_stable_diffusion_completion_fn(model_version: str, **kwargs):
    return ["/mock_data/images/19th_century_wombat_gentleman.png"]


def mock_qdrant_fn(**kwargs):
    from qdrant_client.conversions.common_types import ScoredPoint

    return [
        ScoredPoint(
            id="cf515c6f-1d95-4a21-9052-1a2ecdab34b3",
            version=13,
            score=0.7435235231239,
            payload={"document": "The first president of the United States was George Washington."},
            vector=[0.1, 0.2, 0.3],
        )
    ]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from collections import defaultdict
from typing import Callable, Dict, List, Optional, Type
import logging

from prompttools.prompttest.threshold_type import ThresholdType
from prompttools.prompttest.error.failure import log_failure
from prompttools.experiment import Experiment
from prompttools.prompttest.error.failure import PromptTestSetupException


class PromptTestRunner:
    r"""
    Base class for prompt test runners. Please use the subclass instead.s
    """

    def __init__(self):
        self.ran = defaultdict(bool)
        self.experiments = dict()

    def run(self, *args, **kwargs) -> str:
        r"""
        Runs the test if it has not already been run.
        """
        key = str(args)
        if self.ran[key]:
            return key
        self.experiments[key] = self._get_experiment(*args, **kwargs)
        self.experiments[key].run()
        self.ran[key] = True
        return key

    def evaluate(
        self,
        key: str,
        metric_name: str,
        eval_fn: Callable,
        expected: Optional[str] = None,
    ) -> None:
        r"""
        Evaluates the test results using the given ``eval_fn``.
        """
        self.experiments[key].evaluate(metric_name, eval_fn, expected=expected)

    def visualize(self, key: str) -> None:
        r"""
        Evaluates the test results using the given ``eval_fn``.
        """
        self.experiments[key].visualize()

    def scores(self, key):
        r"""
        Returns the scores for the underlying experiment at the
        given key.
        """
        return self.experiments[key].scores

    @staticmethod
    def _get_experiment(
        experiment: Type[Experiment],
        model_name: str,
        prompts: List[str],
        model_args: Dict[str, object],
    ) -> Experiment:
        return experiment([model_name], prompts, **{k: [v] for k, v in model_args})


prompt_test_runner = PromptTestRunner()


def run_prompttest(
    metric_name: str,
    eval_fn: Callable,
    threshold: float,
    threshold_type: ThresholdType,
    prompts: List[str],
    results: List[str],
    expected: Optional[List[str]],
) -> int:
    """
    Runs the prompt test evaluation.
    """
    scores = []
    for i, result in enumerate(results):
        if expected:
            score = eval_fn(prompts[i], result, metadata={}, expected=expected[i])
        else:
            score = eval_fn(prompts[i], result, metadata={})
        scores.append(score)
    if not scores:
        logging.error("Something went wrong during testing. Make sure your API keys are set correctly.")
        raise PromptTestSetupException
    for score in scores:
        if not (score <= threshold if threshold_type == ThresholdType.MAXIMUM else score >= threshold):
            log_failure(metric_name, threshold, score, threshold_type)
            return 1
    return 0



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

try:
    import google.generativeai as palm
except ImportError:
    palm = None

from prompttools.selector.prompt_selector import PromptSelector
from prompttools.mock.mock import mock_palm_completion_fn
from .experiment import Experiment
from typing import Optional, Union, Iterable
import os


class GooglePaLMCompletionExperiment(Experiment):
    r"""
    This class defines an experiment for Google PaLM's generate text API. It accepts lists for each argument
    passed into PaLM's API, then creates a cartesian product of those arguments, and gets results for each.

    Note:
        - All arguments here should be a ``list``, even if you want to keep the argument frozen
          (i.e. ``temperature=[1.0]``), because the experiment will try all possible combination
          of the input arguments.
        - You should set ``os.environ["GOOGLE_PALM_API_KEY"] = YOUR_KEY`` in order to connect with PaLM's API.

    Args:
        model (list[str]): Which model to call, as a string or a ``types.Model`` (e.g. ``'models/text-bison-001'``).

        prompt (list[str]): Free-form input text given to the model. Given a prompt, the model will
            generate text that completes the input text.

        temperature (list[float]): Controls the randomness of the output. Must be positive.
            Typical values are in the range: ``[0.0, 1.0]``. Higher values produce a
            more random and varied response. A temperature of zero will be deterministic.

        candidate_count (list[int]): The **maximum** number of generated response messages to return.
            This value must be between ``[1, 8]``, inclusive. If unset, this will default to ``1``.

        max_output_tokens (list[int]): Maximum number of tokens to include in a candidate. Must be greater
            than zero. If unset, will default to ``64``.

        top_k (list[float]): The API uses combined nucleus and top-k sampling.
            ``top_k`` sets the maximum number of tokens to sample from on each step.

        top_p (list[float]): The API uses combined nucleus and top-k sampling. ``top_p`` configures the nucleus
            sampling. It sets the maximum cumulative probability of tokens to sample from.

        safety_settings (list[Iterable[palm.types.SafetySettingDict]]): A list of unique ``types.SafetySetting``
            instances for blocking unsafe content.

        stop_sequences (list[Union[str, Iterable[str]]]): A set of up to 5 character sequences that will stop output
            generation. If specified, the API will stop at the first appearance of a stop sequence.
    """

    def __init__(
        self,
        model: list[str],
        prompt: list[str],
        temperature: list[Optional[float]] = [None],
        candidate_count: list[Optional[int]] = [None],
        max_output_tokens: list[Optional[int]] = [None],
        top_p: list[Optional[float]] = [None],
        top_k: list[Optional[float]] = [None],
        safety_settings: list[Optional[Iterable["palm.types.SafetySettingDict"]]] = [None],
        stop_sequences: list[Union[str, Iterable[str]]] = [None],
    ):
        if palm is None:
            raise ModuleNotFoundError(
                "Package `google.generativeai` is required to be installed to use PaLM API in this experiment."
                "Please use `pip install google.generativeai` to install the package"
            )
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_palm_completion_fn()
        else:
            self.completion_fn = self.palm_completion_fn
            palm.configure(api_key=os.environ["GOOGLE_PALM_API_KEY"])

        # If we are using a prompt selector, we need to
        # render the prompts from the selector
        if isinstance(prompt[0], PromptSelector):
            prompt = [selector.for_palm() for selector in prompt]

        self.all_args = dict(
            model=model,
            prompt=prompt,
            temperature=temperature,
            candidate_count=candidate_count,
            max_output_tokens=max_output_tokens,
            top_p=top_p,
            top_k=top_k,
            safety_settings=safety_settings,
            stop_sequences=stop_sequences,
        )
        super().__init__()

    def palm_completion_fn(self, **input_args):
        return palm.generate_text(**input_args)

    @staticmethod
    def _extract_responses(completion_response: "palm.text.text_types.Completion") -> list[str]:
        # `# completion_response.result` will return the top response
        return [candidate["output"] for candidate in completion_response.candidates][0]

    def _get_model_names(self):
        return [combo["model"] for combo in self.argument_combos]

    def _get_prompts(self):
        return [combo["prompt"] for combo in self.argument_combos]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

try:
    from vertexai.preview.language_models import ChatModel, InputOutputTextPair
except ImportError:
    ChatModel = None
    InputOutputTextPair = None

from .experiment import Experiment
from typing import Optional
import copy


class GoogleVertexChatCompletionExperiment(Experiment):
    r"""
    This class defines an experiment for Google Vertex AI's chat API. It accepts lists for each argument
    passed into Vertex AI's API, then creates a cartesian product of those arguments, and gets results for each.

    Note:
        - All arguments here should be a ``list``, even if you want to keep the argument frozen
          (i.e. ``temperature=[1.0]``), because the experiment will try all possible combination
          of the input arguments.
        - You need to set up your Google Vertex AI credentials properly before executing this experiment. One option
          is to execute on Google Cloud's Colab.

    Args:
        model (list[str]): Which model to call, as a string or a ``types.Model`` (e.g. ``'models/text-bison-001'``).

        message (list[str]): Message for the chat model to respond.

        context (list[str]): Context shapes how the model responds throughout the conversation. For example,
            you can use context to specify words the model can or cannot use,
            topics to focus on or avoid, or the response format or style.

        examples (list[list['InputOutputTextPair']]): Examples for the model to learn how to
            respond to the conversation.

        temperature (list[float]): Controls the randomness of the output. Must be positive.
            Typical values are in the range: ``[0.0, 1.0]``. Higher values produce a
            more random and varied response. A temperature of zero will be deterministic.

        max_output_tokens (list[int]): Maximum number of tokens to include in a candidate. Must be greater
            than zero. If unset, will default to ``64``.

        top_k (list[float]): The API uses combined nucleus and top-k sampling.
            ``top_k`` sets the maximum number of tokens to sample from on each step.

        top_p (list[float]): The API uses combined nucleus and top-k sampling. ``top_p`` configures the nucleus
            sampling. It sets the maximum cumulative probability of tokens to sample from.

        stop_sequences (list[Union[str, Iterable[str]]]): A set of up to 5 character sequences that will stop output
            generation. If specified, the API will stop at the first appearance of a stop sequence.
    """

    def __init__(
        self,
        model: list[str],
        message: list[str],
        context: list[Optional[str]] = [None],
        examples: list[Optional[list[InputOutputTextPair]]] = [None],
        temperature: list[Optional[float]] = [None],
        max_output_tokens: list[Optional[int]] = [None],
        top_p: list[Optional[float]] = [None],
        top_k: list[Optional[int]] = [None],
        stop_sequences: list[list[str]] = [None],
    ):
        if ChatModel is None:
            raise ModuleNotFoundError(
                "Package `vertexai` is required to be installed to use Google Vertex API in this experiment."
                "Please use `pip install google-cloud-aiplatform` to install the package"
            )

        self.completion_fn = self.vertex_chat_completion_fn

        self.all_args = dict(
            model=model,
            message=message,
            context=context,
            examples=examples,
            temperature=temperature,
            max_output_tokens=max_output_tokens,
            top_p=top_p,
            top_k=top_k,
            stop_sequences=stop_sequences,
        )
        super().__init__()

    def vertex_chat_completion_fn(self, **input_args):
        chat_model = ChatModel.from_pretrained(model_name=input_args["model"])
        message = input_args["message"]
        params = copy.deepcopy(input_args)
        del params["model"], params["message"]
        chat = chat_model.start_chat(**params)
        return chat.send_message(message)

    @staticmethod
    def _extract_responses(response) -> list[str]:
        # `response.text` will return the top response
        return response.text

    def _get_model_names(self):
        return [combo["model"] for combo in self.argument_combos]

    def _get_prompts(self):
        return [combo["message"] for combo in self.argument_combos]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import os
import openai
import jinja2
import pandas
from .error import PromptToolsUtilityError

EVALUATION_SYSTEM_PROMPT = """
You are a grader evaluating responses to math questions.
Given the PROMPT and EXPECTED, evaluate the ACTUAL answer.
The ACTUAL answer should be the same as the EXPECTED.
You should grade the response as either RIGHT or WRONG.
If the ACTUAL answer is the same as the EXPECTED, mark it RIGHT.
Otherwise, mark it WRONG.
"""

EVALUATION_USER_TEMPLATE = """
PROMPT: {{prompt}}
EXPECTED: {{expected}}
ACTUAL: {{actual}}
ANSWER:
"""


def _get_messages(prompt: str, expected: str, response: str):
    environment = jinja2.Environment()
    template = environment.from_string(EVALUATION_USER_TEMPLATE)
    user_message = template.render({"prompt": prompt, "expected": expected, "actual": response})
    return [
        {"role": "system", "content": EVALUATION_SYSTEM_PROMPT},
        {"role": "user", "content": user_message},
    ]


def compute(prompt: str, expected: str, response: str, model: str = "gpt-4") -> float:
    r"""
    Uses a high quality chat model, like GPT-4, to automatically evaluate a given
    prompt/response pair. Outputs can be 0 or 1.

    Args:
        prompt (str): The input prompt.
        response (str): The model response.
        model (str): The OpenAI chat model to use for generating an expected response.
            Defaults to GPT-4.
    """
    if not os.environ["OPENAI_API_KEY"]:
        raise PromptToolsUtilityError("Missing API key for evaluation.")
    evaluation = openai.chat.completions.create(model=model, messages=_get_messages(prompt, expected, response))
    return 1.0 if "RIGHT" in evaluation.choices[0].message.content else 0.0


def evaluate(prompt: str, response: str, metadata: dict, expected: str) -> float:
    r"""
    Uses auto-evaluation to score the model response.
    """
    return compute(prompt, expected, response)


def autoeval_from_expected_response(
    row: pandas.core.series.Series, expected: str, prompt_column_name: str, response_column_name: str = "response"
):
    prompt = row[prompt_column_name]
    response = row[response_column_name]
    return compute(prompt, expected, response)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.
import itertools
import warnings

import pandas as pd
from typing import Callable, Optional

try:
    import lancedb
    from lancedb.embeddings import with_embeddings

except ImportError:
    lancedb = None

import logging
from time import perf_counter
from .experiment import Experiment
from ._utils import _get_dynamic_columns

VALID_TASKS = [""]


def query_builder(
    table: "lancedb.Table",
    embed_fn: Callable,
    text: str,
    metric: str = "cosine",
    limit: int = 3,
    filter: str = None,
    nprobes: int = None,
    refine_factor: int = None,
):
    if nprobes is not None or refine_factor is not None:
        warnings.warn(
            "`nprobes` and `refine_factor` are not used by the default `query_builder`. "
            "Feel free to open an issue to request adding support for them."
        )
    query = table.search(embed_fn(text)[0]).metric(metric)
    if filter:
        query = query.where(filter)
    return query.limit(limit).to_df()


class LanceDBExperiment(Experiment):
    r"""
    Perform an experiment with ``LanceDB`` to test different embedding functions or retrieval arguments.
    You can query from an existing table, or create a new one (and insert documents into it) during
    the experiment.

    Args:
        uri (str): LanceDB uri to interact with your database. Default is "lancedb"
        table_name (str): the table that you will get or create. Default is "table"
        use_existing_table (bool): determines whether to create a new collection or use
            an existing one
        embedding_fns (list[Callable]): embedding functions to test in the experiment
            by default only uses the default one in LanceDB
        query_args (dict[str, list]): parameters used to query the table
            Each value is expected to be a list to create all possible combinations
        data (Optional[list[dict]]): documents or embeddings that will be added to
            the newly created table
        text_col_name (str): name of the text column in the table. Default is "text"
        clean_up (bool): determines whether to drop the table after the experiment ends
    """

    def __init__(
        self,
        embedding_fns: dict[str, Callable],
        query_args: dict[str, list],
        uri: str = "lancedb",
        table_name: str = "table",
        use_existing_table: bool = False,
        data: Optional[list[dict]] = None,
        text_col_name: str = "text",
        clean_up: bool = False,
    ):
        if lancedb is None:
            raise ModuleNotFoundError(
                "Package `lancedb` is required to be installed to use this experiment."
                "Please use `pip install lancedb` to install the package"
            )
        self.table_name = table_name
        self.use_existing_table = use_existing_table
        self.embedding_fns = embedding_fns
        if use_existing_table and data:
            raise RuntimeError("You can either use an existing collection or create a new one during the experiment.")
        if not use_existing_table and data is None:
            raise RuntimeError("If you choose to create a new collection, you must also add to it.")
        self.data = data if data is not None else []
        self.argument_combos: list[dict] = []
        self.text_col_name = text_col_name
        self.db = lancedb.connect(uri)
        self.completion_fn = self.lancedb_completion_fn
        self.query_args = query_args
        self.clean_up = clean_up
        super().__init__()

    def prepare(self):
        for combo in itertools.product(*self.query_args.values()):
            self.argument_combos.append(dict(zip(self.query_args.keys(), combo)))

    def run(self, runs: int = 1):
        input_args = []  # This will be used to construct DataFrame table
        results = []
        latencies = []
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()

        for emb_fn_name, emb_fn in self.embedding_fns.items():
            if self.use_existing_table:  # Use existing table
                table = self.db.open_table(self.table_name)
                if not table:
                    raise RuntimeError(f"Table {self.table_name} does not exist.")
            else:  # Create table and insert data
                data = with_embeddings(emb_fn, self.data, self.text_col_name)
                table = self.db.create_table(self.table_name, data, mode="overwrite")

            # Query from table
            for query_arg_dict in self.argument_combos:
                query_args = query_arg_dict.copy()
                for _ in range(runs):
                    start = perf_counter()
                    results.append(self.lancedb_completion_fn(table=table, embedding_fn=emb_fn, **query_args))
                    latencies.append(perf_counter() - start)
                    query_args["emb_fn"] = emb_fn_name  # Saving for visualization
                    input_args.append(query_args)

            # Clean up
            if self.clean_up:
                self.db.drop_table(self.table_name)

        self._construct_result_dfs(input_args, results, latencies)

    def lancedb_completion_fn(self, table, embedding_fn, **kwargs):
        return query_builder(table, embedding_fn, **kwargs)

    def _construct_result_dfs(
        self,
        input_args: list[dict[str, object]],
        results: list[dict[str, object]],
        latencies: list[float],
    ):
        r"""
        Construct a few DataFrames that contain all relevant data (i.e. input arguments, results, evaluation metrics).

        This version only extract the most relevant objects returned by LanceDB.

        Args:
             input_args (list[dict[str, object]]): list of dictionaries, where each of them is a set of
                input argument that was passed into the model
             results (list[dict[str, object]]): list of responses from the model
             latencies (list[float]): list of latency measurements
        """
        # `input_arg_df` contains all all input args
        input_arg_df = pd.DataFrame(input_args)
        # `dynamic_input_arg_df` contains input args that has more than one unique values
        dynamic_input_arg_df = _get_dynamic_columns(input_arg_df)

        # `response_df` contains the extracted response (often being the text response)
        response_dict = dict()
        response_dict["top doc ids"] = [self._extract_top_doc_ids(result) for result in results]
        response_dict["distances"] = [self._extract_lancedb_dists(result) for result in results]
        response_dict["documents"] = [self._extract_lancedb_docs(result) for result in results]
        response_df = pd.DataFrame(response_dict)
        # `result_df` contains everything returned by the completion function
        result_df = response_df  # pd.concat([self.response_df, pd.DataFrame(results)], axis=1)

        # `score_df` contains computed metrics (e.g. latency, evaluation metrics)
        self.score_df = pd.DataFrame({"latency": latencies})

        # `partial_df` contains some input arguments, extracted responses, and score
        self.partial_df = pd.concat([dynamic_input_arg_df, response_df, self.score_df], axis=1)
        # `full_df` contains all input arguments, responses, and score
        self.full_df = pd.concat([input_arg_df, result_df, self.score_df], axis=1)

    @staticmethod
    def _extract_top_doc_ids(output: pd.DataFrame) -> list[tuple[str, float]]:
        r"""Helper function to get distances between documents from LanceDB."""
        return output.to_dict(orient="list")["ids"]

    @staticmethod
    def _extract_lancedb_dists(output: pd.DataFrame) -> list[tuple[str, float]]:
        r"""Helper function to get distances between documents from LanceDB."""
        return output.to_dict(orient="list")["_distance"]

    @staticmethod
    def _extract_lancedb_docs(output: pd.DataFrame) -> list[tuple[str, float]]:
        r"""Helper function to get distances between documents from LanceDB."""
        return output.to_dict(orient="list")["text"]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import os
from typing import Dict
import openai
import pandas.core.series
import jinja2
from .error import PromptToolsUtilityError

EVALUATION_SYSTEM_PROMPT = """
Determine whether or not the response is following directions.
Your answer should either be "RIGHT" if the response follows directions,
or "WRONG" if the model is not following directions.
"""

EVALUATION_USER_TEMPLATE = """
PROMPT: {{prompt}}
RESPONSE: {{response}}
ANSWER:
"""


def _get_messages(prompt: str, response: str):
    environment = jinja2.Environment()
    template = environment.from_string(EVALUATION_USER_TEMPLATE)
    user_message = template.render({"prompt": prompt, "response": response})
    return [
        {"role": "system", "content": EVALUATION_SYSTEM_PROMPT},
        {"role": "user", "content": user_message},
    ]


def compute(prompt: str, response: str, model: str = "gpt-4") -> float:
    r"""
    Uses a high quality chat model, like GPT-4, to automatically evaluate a given
    prompt/response pair. Outputs can be 0 or 1.

    Args:
        prompt (str): The input prompt.
        response (str): The model response.
        model (str): The OpenAI chat model to use for generating an expected response.
            Defaults to GPT-4.
    """
    if not os.environ["OPENAI_API_KEY"]:
        raise PromptToolsUtilityError
    evaluation = openai.chat.completions.create(model=model, messages=_get_messages(prompt, response))
    return 1.0 if "RIGHT" in evaluation.choices[0].message.content else 0.0


def evaluate(prompt: str, response: str, _metadata: Dict) -> float:
    r"""
    Uses auto-evaluation to score the model response with "gpt-4" as the judge, returning 0.0 or 1.0.

    Args:
        prompt (str): The input prompt.
        response (str): The model response.
        metadata (str): Not used.
    """
    return compute(prompt, response)


def autoeval_binary_scoring(
    row: pandas.core.series.Series,
    prompt_column_name: str,
    response_column_name: str = "response",
) -> float:
    r"""
    Uses auto-evaluation to score the model response with "gpt-4" as the judge, returning 0.0 or 1.0.

    Args:
        row (pandas.core.series.Series): A row of data from the full DataFrame (including input, model response, other
            metrics, etc).
        prompt_column_name (str): name of the column that contains the input prompt
        response_column_name (str): name of the column that contains the model's response, defaults to ``"response"``
    """
    return compute(row[prompt_column_name], row[response_column_name])



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
import time

import pandas as pd
from typing import Any, Dict, Optional, Iterable

try:
    import pinecone
except ImportError:
    pinecone = None

import itertools
import logging

# from prompttools.mock.mock import mock_chromadb_fn
from time import perf_counter
from .experiment import Experiment
from ._utils import _get_dynamic_columns

VALID_TASKS = [""]


class PineconeExperiment(Experiment):
    r"""
    Perform an experiment with ``Pinecone`` to test different embedding functions or retrieval arguments.
    You can query from an existing collection, or create a new one (and insert documents into it) during
    the experiment. If you choose to create a new collection, it will be automatically cleaned up
    as the experiment ends.

    Args:
        index_name (str): the index that you will use or create
        use_existing_index (bool): determines whether to create a new collection or use
            an existing one
        query_index_params (dict[str, list]): parameters used to query the collection
            Each value is expected to be a list to create all possible combinations
        create_index_params (Optional[dict]): configuration of the new index (e.g. number of dimensions,
            distance function)
        data (Optional[list]): documents or embeddings that will be added to
            the newly created collection
    """

    def __init__(
        self,
        index_name: str,
        use_existing_index: bool,
        query_index_params: dict,
        create_index_params: Optional[dict] = None,
        data: Optional[list] = None,
    ):
        if pinecone is None:
            raise ModuleNotFoundError(
                "Package `pinecone` is required to be installed to use this experiment."
                "Please use `pip install pinecone-client` to install the package"
            )
        pinecone.init(api_key=os.environ["PINECONE_API_KEY"], environment=os.environ["PINECONE_ENVIRONMENT"])
        self.index_name = index_name
        self.completion_fn = self.pinecone_completion_fn
        # if os.getenv("DEBUG", default=False):
        #     self.completion_fn = mock_chromadb_fn
        self.use_existing_index = use_existing_index
        self.create_index_params: dict = create_index_params if create_index_params else {}
        self.data: list = data if data is not None else []
        self.query_index_params = query_index_params
        if use_existing_index and create_index_params:
            raise RuntimeError("You can either use an existing collection or create a new one during the experiment.")
        if not use_existing_index and data is None:
            raise RuntimeError("If you choose to create a new collection, you must also add to it.")
        super().__init__()

    # @classmethod
    # def initialize(cls, test_parameters: dict[str, list], frozen_parameters: dict):
    #     required_frozen_params = (
    #         "chroma_client",
    #         "collection_name",
    #         "use_existing_collection",
    #         "query_collection_params",
    #     )
    #     for arg_name in required_frozen_params:
    #         if arg_name not in frozen_parameters or arg_name in test_parameters:
    #             raise RuntimeError(f"'{arg_name}' must be a frozen parameter in ChromaDBExperiment.")
    #     frozen_parameters = {k: [v] for k, v in frozen_parameters.items()}
    #     return cls(**test_parameters, **frozen_parameters)

    def pinecone_completion_fn(
        self,
        index: "pinecone.Index",
        **query_params: Dict[str, Any],
    ):
        r"""
        Pinecone helper function to make request
        """
        result = index.query(**query_params)
        return result

    def prepare(self) -> None:
        r"""
        Creates argument combinations by taking the cartesian product of all inputs.
        """
        # TODO: Add support for the case where `create_index_params` are a list, add logic to expand
        for combo in itertools.product(*self.query_index_params.values()):
            self.argument_combos.append(dict(zip(self.query_index_params.keys(), combo)))

    @staticmethod
    def _batch_upsert(pinecone_index: "pinecone.Index", data: Iterable) -> None:
        batch = []
        for d in data:
            batch.append(d)
            if len(batch) == 100:
                pinecone_index.upsert(batch)
                batch = []
        pinecone_index.upsert(batch)

    @staticmethod
    def _wait_for_eventual_consistency(pinecone_index: "pinecone.Index", n_sample: int) -> None:
        i = 0
        while pinecone_index.describe_index_stats()["total_vector_count"] < n_sample:
            i += 1
            print("Waiting for Pinecone's eventual consistency after inserting data.")
            time.sleep(3)
            if i == 20:
                raise TimeoutError("Pinecone has not insert data due to eventual consistency after 1 minute.")

    def run(self, runs: int = 1):
        input_args = []  # This will be used to construct DataFrame table
        results = []
        latencies = []
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()

        # Insert data
        if not self.use_existing_index:
            # TODO: Add support for the case where params are a list , add logic to `prepare`
            pinecone.create_index(self.index_name, **self.create_index_params)
            index = pinecone.Index(self.index_name)
            self._batch_upsert(index, self.data)
        else:
            index = pinecone.Index(self.index_name)

        if not self.use_existing_index:
            self._wait_for_eventual_consistency(index, len(self.data))

        # Query
        for query_arg_dict in self.argument_combos:
            arg_combo = query_arg_dict.copy()
            for _ in range(runs):
                input_args.append(arg_combo)
                start = perf_counter()
                results.append(self.pinecone_completion_fn(index, **query_arg_dict))
                latencies.append(perf_counter() - start)

        # Clean up
        if not self.use_existing_index:
            pinecone.delete_index(self.index_name)
        self._construct_result_dfs(input_args, results, latencies)

    def _construct_result_dfs(
        self,
        input_args: list[dict[str, object]],
        results: list[dict[str, object]],
        latencies: list[float],
    ):
        r"""
        Construct a few DataFrames that contain all relevant data (i.e. input arguments, results, evaluation metrics).

        This version only extract the most relevant objects returned by Pinecone.

        Args:
             input_args (list[dict[str, object]]): list of dictionaries, where each of them is a set of
                input argument that was passed into the model
             results (list[dict[str, object]]): list of responses from the model
             latencies (list[float]): list of latency measurements
        """
        # `input_arg_df` contains all all input args
        input_arg_df = pd.DataFrame(input_args)
        # `dynamic_input_arg_df` contains input args that has more than one unique values
        dynamic_input_arg_df = _get_dynamic_columns(input_arg_df)

        # `response_df` contains the extracted response (often being the text response)
        response_dict = dict()
        response_dict["top doc ids"] = [self._extract_top_doc_ids(result) for result in results]
        response_dict["scores"] = [self._extract_pinecone_scores(result) for result in results]
        response_dict["documents"] = [self._extract_pinecone_docs(result) for result in results]
        response_df = pd.DataFrame(response_dict)
        # `result_df` contains everything returned by the completion function
        result_df = response_df  # pd.concat([self.response_df, pd.DataFrame(results)], axis=1)

        # `score_df` contains computed metrics (e.g. latency, evaluation metrics)
        self.score_df = pd.DataFrame({"latency": latencies})

        # `partial_df` contains some input arguments, extracted responses, and score
        self.partial_df = pd.concat([dynamic_input_arg_df, response_df, self.score_df], axis=1)
        # `full_df` contains all input arguments, responses, and score
        self.full_df = pd.concat([input_arg_df, result_df, self.score_df], axis=1)

    @staticmethod
    def _extract_top_doc_ids(output: "pinecone.QueryResponse") -> list[tuple[str, float]]:
        r"""Helper function to get top document IDs from Pinecone."""
        return [match["id"] for match in output["matches"]]

    @staticmethod
    def _extract_pinecone_scores(output: "pinecone.QueryResponse") -> list[tuple[str, float]]:
        r"""Helper function to get the scores of documents from Pinecone."""
        return [match["score"] for match in output["matches"]]

    @staticmethod
    def _extract_pinecone_docs(output: "pinecone.QueryResponse") -> list[tuple[str, float]]:
        r"""Helper function to get top documents from Pinecone."""
        return [match["values"] for match in output["matches"]]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Optional
from prompttools.experiment import Experiment


from prompttools.common import HEGEL_BACKEND_URL
import os
import pickle
import requests


class ExperimentationHarness:
    r"""
    Base class for experimentation harnesses. This should not be used directly, please use the subclasses instead.
    """

    experiment: Experiment
    PIVOT_COLUMNS: list

    def __init__(self) -> None:
        self.input_pairs_dict = None
        self.experiment = None
        self.runs = 1
        self._experiment_id = None
        self._revision_id = None

    @staticmethod
    def _prepare_arguments(arguments: dict[str, object]) -> dict[str, list[object]]:
        return {name: [arg] for name, arg in arguments.items()}

    def prepare(self) -> None:
        r"""
        Prepares the underlying experiment.
        """
        self.experiment.prepare()

    def run(self, clear_previous_results: bool = False) -> None:
        r"""
        Runs the underlying experiment.
        """
        self.experiment.run(runs=self.runs, clear_previous_results=clear_previous_results)

    def evaluate(self, metric_name: str, eval_fn: Callable, static_eval_fn_kwargs: dict = {}, **eval_fn_kwargs) -> None:
        r"""
        Uses the given eval_fn to evaluate the results of the underlying experiment.
        """
        self.experiment.evaluate(metric_name, eval_fn, static_eval_fn_kwargs, **eval_fn_kwargs)

    # def gather_feedback(self) -> None:
    #     self.experiment.gather_feedback(self.input_pairs_dict, self.PIVOT_COLUMNS)

    def visualize(self, pivot: bool = False) -> None:
        r"""
        Displays a visualization of the experiment results.
        """
        if pivot:
            self.experiment.visualize(pivot_columns=self.PIVOT_COLUMNS, pivot=True)
        else:
            self.experiment.visualize()

    def rank(self, metric_name: str, is_average: bool = False) -> dict[str, float]:
        r"""
        Scores and ranks the experiment inputs using the pivot columns,
        e.g. prompt templates or system prompts.
        """
        return self.experiment.rank(self.input_pairs_dict, self.PIVOT_COLUMNS, metric_name, is_average)

    @property
    def full_df(self):
        return self.experiment.full_df

    @property
    def partial_df(self):
        return self.experiment.partial_df

    @property
    def score_df(self):
        return self.experiment.score_df

    def _get_state(self):
        raise NotImplementedError("Should be implemented by specific harness class.")

    @classmethod
    def _load_state(cls, state, experiment_id: str, revision_id: str, experiment_type_str: str):
        raise NotImplementedError("Should be implemented by specific harness class.")

    def save_experiment(self, name: Optional[str] = None):
        r"""
        name (str, optional): Name of the experiment. This is optional if you have previously loaded an experiment
            into this object.
        """
        if name is None and self._experiment_id is None:
            raise RuntimeError("Please provide a name for your experiment.")
        if self.full_df is None:
            raise RuntimeError("Cannot save empty experiment. Please run it first.")
        if os.environ["HEGELAI_API_KEY"] is None:
            raise PermissionError("Please set HEGELAI_API_KEY (e.g. os.environ['HEGELAI_API_KEY']).")
        state = self._get_state()
        url = f"{HEGEL_BACKEND_URL}/sdk/save"
        headers = {
            "Content-Type": "application/octet-stream",  # Use a binary content type for pickled data
            "Authorization": os.environ["HEGELAI_API_KEY"],
        }
        print("Sending HTTP POST request...")
        data = pickle.dumps((name, self._experiment_id, self._experiment_type, state))
        response = requests.post(url, data=data, headers=headers)
        self._experiment_id = response.json().get("experiment_id")
        self._revision_id = response.json().get("revision_id")
        return response

    @classmethod
    def load_experiment(cls, experiment_id: str):
        r"""
        experiment_id (str): experiment ID of the experiment that you wish to load.
        """
        if os.environ["HEGELAI_API_KEY"] is None:
            raise PermissionError("Please set HEGELAI_API_KEY (e.g. os.environ['HEGELAI_API_KEY']).")

        url = f"{HEGEL_BACKEND_URL}/sdk/get/experiment/{experiment_id}"
        headers = {
            "Content-Type": "application/octet-stream",  # Use a binary content type for pickled data
            "Authorization": os.environ["HEGELAI_API_KEY"],
        }
        print("Sending HTTP GET request...")
        response = requests.get(url, headers=headers)
        if response.status_code == 200:  # Note that state should not have `name` included
            new_experiment_id, revision_id, experiment_type_str, state = pickle.loads(response.content)
            if new_experiment_id != experiment_id:
                raise RuntimeError("Experiment ID mismatch between request and response.")
            return cls._load_state(state, experiment_id, revision_id, experiment_type_str)
        else:
            print(f"Error: {response.status_code}, {response.text}")

    @classmethod
    def load_revision(cls, revision_id: str):
        r"""
        revision_id (str): revision ID of the experiment that you wish to load.
        """
        if os.environ["HEGELAI_API_KEY"] is None:
            raise PermissionError("Please set HEGELAI_API_KEY (e.g. os.environ['HEGELAI_API_KEY']).")

        url = f"{HEGEL_BACKEND_URL}/sdk/get/revision/{revision_id}"
        headers = {
            "Content-Type": "application/octet-stream",  # Use a binary content type for pickled data
            "Authorization": os.environ["HEGELAI_API_KEY"],
        }
        print("Sending HTTP GET request...")
        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            experiment_id, new_revision_id, experiment_type_str, state = pickle.loads(response.content)
            if new_revision_id != revision_id:
                raise RuntimeError("Revision ID mismatch between request and response.")
            return cls._load_state(state, experiment_id, revision_id, experiment_type_str)
        else:
            print(f"Error: {response.status_code}, {response.text}")



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


from .harness import ExperimentationHarness
from .chat_history_harness import ChatHistoryExperimentationHarness
from .chat_model_comparison_harness import ChatModelComparisonHarness
from .chat_prompt_template_harness import ChatPromptTemplateExperimentationHarness
from .model_comparison_harness import ModelComparisonHarness
from .multi_experiment_harness import MultiExperimentHarness
from .prompt_template_harness import PromptTemplateExperimentationHarness
from .rag_harness import RetrievalAugmentedGenerationExperimentationHarness
from .system_prompt_harness import SystemPromptExperimentationHarness


__all__ = [
    "ChatHistoryExperimentationHarness",
    "ChatModelComparisonHarness",
    "ChatPromptTemplateExperimentationHarness",
    "ExperimentationHarness",
    "ModelComparisonHarness",
    "MultiExperimentHarness",
    "PromptTemplateExperimentationHarness",
    "RetrievalAugmentedGenerationExperimentationHarness",
    "SystemPromptExperimentationHarness",
]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the


# LICENSE file in the root directory of this source tree.

import pandas as pd
from typing import Callable, Optional

try:
    import weaviate
except ImportError:
    weaviate = None
import logging

from time import perf_counter
from .experiment import Experiment
from ._utils import _get_dynamic_columns

VALID_TASKS = [""]


def default_query_builder(
    client: "weaviate.Client",
    class_name: str,
    property_names: list[str],
    text_query: str,
):
    near_text_search_operator = {"concepts": [text_query]}
    return client.query.get(class_name, property_names).with_near_text(near_text_search_operator).with_limit(limit=3)


class WeaviateExperiment(Experiment):
    r"""
    Perform an experiment with Weaviate to test different vectorizers or querying functions.
    You can query from an existing class, or create a new one (and insert data objects into it) during
    the experiment. If you choose to create a new class, it will be automatically cleaned up
    as the experiment ends.

    Args:
        client (weaviate.Client): The Weaviate client instance to interact with the Weaviate server.
        class_name (str): The name of the Weaviate class (equivalent to a collection in ChromaDB).
        use_existing_data (bool): If ``True``, indicates that existing data will be used for the experiment.
            If ``False``, new data objects will be inserted into Weaviate during the experiment.
        property_names (list[str]): List of property names in the Weaviate class to be used in the experiment.
        text_queries (list[str]): List of text queries to be used for retrieval in the experiment.
        query_builders (dict[str, Callable], optional): A dictionary containing different query builders.
            The key should be the name of the function for visualization purposes.
            The value should be a Callable function that constructs and returns a Weaviate query object.
            Defaults to a built-in query function.
        vectorizers_and_moduleConfigs (Optional[list[tuple[str, dict]]], optional): List of tuples, where each tuple
            contains the name of the vectorizer and its corresponding moduleConfig as a dictionary. This
            is used during data insertion (if necessary).
        property_definitions (Optional[list[dict]], optional): List of property definitions for the Weaviate class.
            Each property definition is a dictionary containing the property name and data type.
            This is used during data insertion (if necessary).
        data_objects (Optional[list], optional): List of data objects to be inserted into Weaviate during the
            experiment. Each data object is a dictionary representing the property-value pairs.
        distance_metrics (Optional[list[str]], optional): List of distance metrics to be used in the experiment.
            These metrics will be used for generating vectorIndexConfig. This is used to define the class object.
            If necessary, either use ``distance_metrics`` or ``vectorIndexConfigs``, not both.
        vectorIndexConfigs (Optional[list[dict]], optional): List of vectorIndexConfig to be used in the
            experiment to define the class object.

    Note:
        - If ``use_existing_data`` is ``False``, the experiment will create a new Weaviate class and insert
          ``data_objects`` into it. The class and ``data_objects`` will be automatically cleaned up at the end of the
          experiment.
        - Either use existing data or specify ``data_objs`` and ``vectorizers`` for insertion.
        - Either ``distance_metrics`` or ``vectorIndexConfigs`` should be provided if necessary, not both.
        - If you pass in a custom ``query_builder`` function, it should accept the same parameters as
          `the default one as seen here`_.

    .. _the default one as seen here:
        https://github.com/hegelai/prompttools/blob/main/prompttools/experiment/experiments/weaviate_experiment.py
    """

    def __init__(
        self,
        client: "weaviate.Client",
        class_name: str,
        use_existing_data: bool,
        property_names: list[str],
        text_queries: list[str],
        query_builders: dict[str, Callable] = {"default": default_query_builder},
        vectorizers_and_moduleConfigs: Optional[list[tuple[str, dict]]] = None,
        property_definitions: Optional[list[dict]] = None,
        data_objects: Optional[list] = None,
        distance_metrics: Optional[list[str]] = None,
        vectorIndexConfigs: Optional[list[dict]] = None,
    ):
        if weaviate is None:
            raise ModuleNotFoundError(
                "Package `weaviate` is required to be installed to use this experiment."
                "Please use `pip install weaviate-client` to install the package"
            )
        self.client = client
        self.completion_fn = self.weaviate_completion_fn
        # TODO: Add a mock function if needed
        # if os.getenv("DEBUG", default=False):
        #     self.completion_fn = mock_chromadb_fn
        self.class_name = class_name
        self.vectorizers_and_moduleConfigs = vectorizers_and_moduleConfigs
        self.property_names = property_names
        self.property_definitions = property_definitions
        if distance_metrics and vectorIndexConfigs:
            raise RuntimeError("Either use `distance_metrics` or `vectorIndexConfigs`.")
        if use_existing_data and data_objects:
            raise RuntimeError("Either use existing data or do not specify `data_objs` for insertion.")
        if not use_existing_data and not data_objects:
            raise RuntimeError("Either use existing data or specify `data_objs` for insertion.")
        if not use_existing_data and not vectorizers_and_moduleConfigs:
            raise RuntimeError("Either use existing data or specify `vectorizers_and_moduleConfigs` for insertion.")
        self.use_existing_data = use_existing_data
        self.is_custom_vectorIndexConfigs = vectorIndexConfigs or distance_metrics
        if vectorIndexConfigs:
            self.vectorIndexConfigs = vectorIndexConfigs
        elif distance_metrics:
            self.vectorIndexConfigs = [
                self._generate_vectorIndexConfigs(dist_metric) for dist_metric in distance_metrics
            ]
        else:  # weaviate's default
            self.vectorIndexConfigs = [self._generate_vectorIndexConfigs("cosine")]
        self.data_objects = data_objects
        self.text_queries = text_queries
        self.query_builders = query_builders
        super().__init__()

    @classmethod
    def initialize(cls, test_parameters: dict[str, list], frozen_parameters: dict):
        required_frozen_params = ("client", "class_name", "use_existing_data")
        for arg_name in required_frozen_params:
            if arg_name not in frozen_parameters or arg_name in test_parameters:
                raise RuntimeError(f"'{arg_name}' must be a frozen parameter in WeaviateExperiment.")
        frozen_parameters = {k: [v] for k, v in frozen_parameters.items()}
        return cls(**test_parameters, **frozen_parameters)

    @staticmethod
    def _generate_vectorIndexConfigs(distance_metric: str):
        return {
            "vectorIndexConfig": {
                "distance": distance_metric,
            }
        }

    def weaviate_completion_fn(
        self,
        query_builder: Callable,
        text_query: str,
    ):
        r"""
        Weaviate helper function to make request
        """
        query_obj = query_builder(self.client, self.class_name, self.property_names, text_query)
        return query_obj.do()

    def prepare(self) -> None:
        r"""
        Creates argument combinations by taking the cartesian product of all inputs.
        """
        self.argument_combos: list[dict] = []
        vectorizers_and_moduleConfigs = (
            [("", {})] if self.vectorizers_and_moduleConfigs is None else self.vectorizers_and_moduleConfigs
        )
        for vectorizer, moduleConfig in vectorizers_and_moduleConfigs:
            for vectorIndexConfig in self.vectorIndexConfigs:
                for query_builder_name, query_builder in self.query_builders.items():
                    for text_query in self.text_queries:
                        self.argument_combos.append(
                            {
                                "vectorizer": vectorizer,
                                "moduleConfig": moduleConfig,
                                "vectorIndexConfig": vectorIndexConfig,
                                "text_query": text_query,
                                "query_builder_name": query_builder_name,
                            }
                        )

    def run(self, runs: int = 1):
        results = []
        latencies = []
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        for arg_dict in self.argument_combos:
            if not self.use_existing_data:
                # Create class (equivalent to collection in ChromaDB)
                weaviate_class_obj = {
                    "class": self.class_name,
                    "vectorizer": arg_dict["vectorizer"],
                    "moduleConfig": arg_dict["moduleConfig"],
                    "properties": self.property_definitions,
                    "vectorIndexConfig": arg_dict["vectorIndexConfig"],
                }
                self.client.schema.create_class(weaviate_class_obj)

                # Batch Insert Items
                logging.info("Inserting items into Weaviate...")
                with self.client.batch() as batch:
                    for data_obj in self.data_objects:
                        batch.add_data_object(
                            data_obj,
                            # data_obj: {'property_name1': 'property_value1', 'property_name2': 'property_value2'}
                            self.class_name,
                        )

            # Query
            query_builder = self.query_builders[arg_dict["query_builder_name"]]
            for _ in range(runs):
                start = perf_counter()
                result = self.completion_fn(query_builder, arg_dict["text_query"])
                latencies.append(perf_counter() - start)
                results.append(result)

            # Clean up
            logging.info("Cleaning up items in Weaviate...")
            if not self.use_existing_data:
                self.client.schema.delete_class(self.class_name)
        self._construct_result_dfs([c for c in self.argument_combos for _ in range(runs)], results, latencies)

    def _construct_result_dfs(
        self,
        input_args: list[dict[str, object]],
        results: list[dict[str, object]],
        latencies: list[float],
    ):
        r"""
        Construct a few DataFrames that contain all relevant data (i.e. input arguments, results, evaluation metrics).

        This version only extract the most relevant objects returned by Weaviate.

        Args:
             input_args (list[dict[str, object]]): list of dictionaries, where each of them is a set of
                input argument that was passed into the model
             results (list[dict[str, object]]): list of responses from the model
        """
        # `input_arg_df` contains all all input args
        input_arg_df = pd.DataFrame(input_args)
        # `dynamic_input_arg_df` contains input args that has more than one unique values
        dynamic_input_arg_df = _get_dynamic_columns(input_arg_df)

        # `response_df` contains the extracted response (often being the text response)
        response_dict = dict()
        response_dict["top objs"] = [self._extract_responses(result) for result in results]
        response_df = pd.DataFrame(response_dict)
        # `result_df` contains everything returned by the completion function
        result_df = response_df  # pd.concat([self.response_df, pd.DataFrame(results)], axis=1)

        # `score_df` contains computed metrics (e.g. latency, evaluation metrics)
        self.score_df = pd.DataFrame({"latency": latencies})

        # `partial_df` contains some input arguments, extracted responses, and score
        self.partial_df = pd.concat([dynamic_input_arg_df, response_df, self.score_df], axis=1)
        # `full_df` contains all input arguments, responses, and score
        self.full_df = pd.concat([input_arg_df, result_df, self.score_df], axis=1)

    def _extract_responses(self, response: dict) -> list[dict]:
        return response["data"]["Get"][self.class_name]



import itertools
import json
import logging
import os
import time
import warnings
from collections import defaultdict
from typing import Callable, List, Iterable, Optional, Dict, Any

from prompttools.experiment.experiments.error import PromptExperimentException
from prompttools.mock.mock import mock_qdrant_fn

try:
    import qdrant_client
except ImportError:
    qdrant_client = None

from prompttools.experiment import Experiment

VALID_TASKS = [""]


Embedding = List[float]
EmbeddingFn = Callable[[str], Embedding]


class QdrantExperiment(Experiment):
    DEFAULT_DISTANCE = "Cosine"

    def __init__(
        self,
        client: "qdrant_client.QdrantClient",
        collection_name: str,
        embedding_fn: EmbeddingFn,
        vector_size: int,
        documents: Iterable[str],
        queries: Iterable[str],
        collection_params: Optional[Dict[str, List[Any]]] = None,
        query_params: Optional[Dict[str, List[Any]]] = None,
    ):
        if qdrant_client is None:
            raise ModuleNotFoundError(
                "Package `qdrant-client` is required to be installed to use this "
                "experiment. Please use `pip install qdrant-client` to install the "
                "package"
            )
        self.client = client
        self.collection_name = collection_name
        self.embedding_fn = embedding_fn
        self.documents = documents
        self.queries = queries
        self.collection_params = collection_params or {}

        if "vectors_config__size" in collection_params:
            warnings.warn(
                "The parameter 'vectors_config__size' is not allowed in "
                "QdrantExperiment. The vector size is determined by the embedding "
                "function. It will be overwritten by {}".format(vector_size)
            )
        # The vector size is a required parameter for Qdrant and has to be overwritten
        self.collection_params["vectors_config__size"] = [vector_size]

        # The distance is set only if not provided by the user
        if "vectors_config__distance" not in collection_params:
            self.collection_params["vectors_config__distance"] = [self.DEFAULT_DISTANCE]

        self.query_params = query_params or {}
        self.completion_fn = self.qdrant_completion_fn
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_qdrant_fn
        self.collection_args_combo: List[dict] = []
        self.query_argument_combos: List[dict] = []
        self.vectorized_documents: List["qdrant_client.models.Record"] = []
        super().__init__()

    @classmethod
    def initialize(cls, test_parameters: dict[str, dict[str, list]], frozen_parameters: dict):
        required_frozen_params = (
            "client",
            "collection_name",
            "embedding_fn",
            "vector_size",
            "documents",
        )
        for arg_name in required_frozen_params:
            if arg_name not in frozen_parameters or arg_name in test_parameters:
                raise RuntimeError(f"'{arg_name}' must be a frozen parameter in QdrantExperiment.")
        return cls(**test_parameters, **frozen_parameters)

    def qdrant_completion_fn(self, **kwargs) -> List["qdrant_client.qdrant_client.types.ScoredPoint"]:
        query_result = self.client.search(self.collection_name, with_vectors=True, with_payload=True, **kwargs)
        return query_result

    def prepare(self) -> None:
        from qdrant_client import models

        # Vectorize the documents and queries
        self.vectorized_documents = [
            models.Record(
                id=i,
                vector=self.embedding_fn(document),
                payload={"document": document},
            )
            for i, document in enumerate(self.documents)
        ]
        self.query_params["query_vector"] = [self.embedding_fn(query) for query in self.queries]

        self.collection_args_combo: list[dict] = []
        for combo in itertools.product(*self.collection_params.values()):
            self.collection_args_combo.append(dict(zip(self.collection_params.keys(), combo)))

        self.query_argument_combos: list[dict] = []
        for combo in itertools.product(*self.query_params.values()):
            self.query_argument_combos.append(dict(zip(self.query_params.keys(), combo)))

    def run(self, runs: int = 1) -> None:
        from qdrant_client import models

        input_args, results, latencies = [], [], []
        if not self.query_argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        for collection_args in self.collection_args_combo:
            collection_args = self._create_nested_object(collection_args)

            try:
                # Create the collection
                self.client.create_collection(self.collection_name, **collection_args)

                # Upload and index the documents
                self.client.upload_records(self.collection_name, self.vectorized_documents)

                # Wait for the collection to be indexed
                while True:
                    collection_info = self.client.get_collection(self.collection_name)
                    if collection_info.status == models.CollectionStatus.GREEN:
                        logging.info("Collection is indexed")
                        break
                    logging.info("Waiting for collection to be indexed...")
                    time.sleep(1)

                # Run the queries
                for query_args in self.query_argument_combos:
                    query_args = self._create_nested_object(query_args)
                    for _ in range(runs):
                        input_args.append(query_args)
                        self.queue.enqueue(self.completion_fn, query_args)

                results.extend(self.queue.get_results())
                latencies.extend(self.queue.get_latencies())
            finally:
                self.client.delete_collection(self.collection_name)

        self._construct_result_dfs(input_args, results, latencies)

    @staticmethod
    def _extract_responses(output: List["qdrant_client.qdrant_client.types.ScoredPoint"]) -> list[str]:
        return [response.payload["document"] for response in output]

    def _create_nested_object(self, args: Dict[str, Any]) -> Dict[str, Any]:
        r"""
        Create a nested object using a flat argument dictionary. If a dict key contains
        a double underscore, it is considered a nested object. For example, if the
        argument dictionary contains the key `a__b__c`, this function will return
        a nested object of the form `{"a": {"b": {"c": <value>}}}`.

        If there are multiple keys with the same prefix, the returned object will
        contain a list of nested objects. For example, if the argument dictionary
        contains the keys `a__b__c` and `a__b__d`, this function will return
        a nested object of the form `{"a": {"b": [{"c": <value>}, {"d": <value>}]}}`.

        Args:
            args: A flat argument dictionary

        Returns:
            dict: A nested object
        """
        tree = lambda: defaultdict(tree)

        nested_object = tree()
        for key, value in args.items():
            key_parts = key.split("__")
            current = nested_object
            for key_part in key_parts[:-1]:
                current = current[key_part]
            current[key_parts[-1]] = value

        # Convert defaultdict to dict
        nested_object = json.loads(json.dumps(nested_object))
        return nested_object



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.
import copy
import os
import json
import pickle
from typing import Dict, List, Optional, Union
import openai
import requests
import itertools
import logging

from prompttools.selector.prompt_selector import PromptSelector
from prompttools.mock.mock import mock_openai_chat_completion_fn, mock_openai_chat_function_completion_fn
from .experiment import Experiment
from .error import PromptExperimentException
import pandas as pd
from prompttools.common import HEGEL_BACKEND_URL


class OpenAIChatExperiment(Experiment):
    r"""
    This class defines an experiment for OpenAI's chat completion API.
    It accepts lists for each argument passed into OpenAI's API, then creates
    a cartesian product of those arguments, and gets results for each.

    Note:
        - All arguments here should be a ``list``, even if you want to keep the argument frozen
          (i.e. ``temperature=[1.0]``), because the experiment will try all possible combination
          of the input arguments.
        - For detailed description of the input arguments, please reference at OpenAI's chat completion API.

    Args:
        model (list[str]): list of ID(s) of the model(s) to use, e.g. ``["gpt-3.5-turbo", "ft:gpt-3.5-turbo:org_id"]``
            If you are using Azure OpenAI service, put the models' deployment names here

        messages (list[dict]): A list of messages comprising the conversation so far. Each message is represented as a
            dictionary with the following keys: ``role: str``, ``content: str``.

        temperature (list[float]):
            Defaults to [1.0]. What sampling temperature to use, between 0 and 2. Higher values like 0.8 will make
            the output more random, while lower values like 0.2 will make it more focused and deterministic.

        top_p (list[float]):
            Defaults to [1.0]. An alternative to sampling with temperature, called nucleus sampling, where the
            model considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens
            comprising the top 10% probability mass are considered.

        n (list[int]):
            Defaults to [1]. How many chat completion choices to generate for each input message.

        stream (list[bool]):
            Defaults to [False]. If set, partial message deltas will be sent, like in ChatGPT. Tokens will be sent
            as data-only server-sent events as they become available, with the stream terminated by a data: [DONE]
            message.

        stop (list[list[str]]):
            Defaults to [None]. Up to 4 sequences where the API will stop generating further tokens.

        max_tokens (list[int]):
            Defaults to [inf]. The maximum number of tokens to generate in the chat completion.

        presence_penalty (list[float]):
            Defaults to [0.0]. Number between -2.0 and 2.0. Positive values penalize new tokens based on whether
            they appear in the text so far, increasing the model's likelihood to talk about new topics.

        frequency_penalty (list[float]):
            Defaults to [0.0]. Number between -2.0 and 2.0. Positive values penalize new tokens based on their
            existing frequency in the text so far, decreasing the model's likelihood to repeat the same line
            verbatim.

        logit_bias (list[dict]):
            Defaults to [None]. Modify the likelihood of specified tokens appearing in the completion. Accepts a
            json object that maps tokens (specified by their token ID in the tokenizer) to an associated bias value
            from -100 to 100.

        functions (list[dict]):
            Defaults to [None]. A list of dictionaries, each of which contains the definition of a function
            the model may generate JSON inputs for.

        function_call (list[dict]):
            Defaults to [None]. A dictionary containing the name and arguments of a function that should be called,
            s generated by the model.

        response_format (list[Optional[dict]]):
            Setting to `{ type: "json_object" }` enables JSON mode, which guarantees the message
            the model generates is valid JSON.

        seed (list[Optional[int]]):
            This feature is in Beta. If specified, our system will make a best effort to sample deterministically,
            such that repeated requests with the same `seed` and parameters should return the same result.
            Determinism is not guaranteed, and you should refer to the `system_fingerprint` response parameter to
            monitor changes in the backend.

        azure_openai_service_configs (Optional[dict]):
            Defaults to ``None``. If it is set, the experiment will use Azure OpenAI Service. The input dict should
            contain these 2 keys (but with values based on your use case and configuration):
            ``{"AZURE_OPENAI_ENDPOINT": "https://YOUR_RESOURCE_NAME.openai.azure.com/", "API_VERSION": "2023-05-15"}``
    """

    _experiment_type = "RawExperiment"

    def __init__(
        self,
        model: List[str] = ["gpt-3.5-turbo"],
        messages: Union[List[List[Dict[str, str]]], List[PromptSelector]] = [],
        temperature: Optional[List[float]] = [1.0],
        top_p: Optional[List[float]] = [1.0],
        n: Optional[List[int]] = [1],
        stream: Optional[List[bool]] = [False],
        stop: Optional[List[List[str]]] = [None],
        max_tokens: Optional[List[int]] = [float("inf")],
        presence_penalty: Optional[List[float]] = [0.0],
        frequency_penalty: Optional[List[float]] = [0.0],
        logit_bias: Optional[List[Dict]] = [None],
        response_format: List[Optional[Dict]] = [None],
        seed: List[Optional[int]] = [None],
        functions: Optional[List[Dict]] = [None],
        function_call: Optional[List[Dict[str, str]]] = [None],
        azure_openai_service_configs: Optional[dict] = None,
    ):
        if azure_openai_service_configs is None:
            self.completion_fn = openai.chat.completions.create
        else:
            client = openai.AzureOpenAI(
                api_key=os.environ["AZURE_OPENAI_KEY"],
                api_version=azure_openai_service_configs["API_VERSION"],
                azure_endpoint=azure_openai_service_configs["AZURE_OPENAI_ENDPOINT"],
            )
            self.completion_fn = client.chat.completions.create
        if os.getenv("DEBUG", default=False):
            if functions[0] is not None:
                self.completion_fn = mock_openai_chat_function_completion_fn
            else:
                self.completion_fn = mock_openai_chat_completion_fn

        # If we are using a prompt selector, we need to render
        # messages, as well as create prompt_keys to map the messages
        # to corresponding prompts in other models.
        if len(messages) > 0 and isinstance(messages[0], PromptSelector):
            self.prompt_keys = {
                str(selector.for_openai_chat()[-1]["content"]): selector.for_llama() for selector in messages
            }
            messages = [selector.for_openai_chat() for selector in messages]
        else:
            self.prompt_keys = messages

        self.all_args = dict(
            model=model,
            messages=messages,
            temperature=temperature,
            functions=functions,
            function_call=function_call,
            top_p=top_p,
            n=n,
            stream=stream,
            stop=stop,
            max_tokens=max_tokens,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            seed=seed,
            response_format=response_format,
        )

        # These parameters aren't supported by `gpt-35-turbo`, we can remove them if they are equal to defaults
        # This has no impact on the default case
        if self.all_args["logit_bias"] == [None]:
            del self.all_args["logit_bias"]

        super().__init__()

    @staticmethod
    def _extract_responses(output: openai.types.Completion) -> str:
        message = output.choices[0].message
        if hasattr(message, "function_call") and message.function_call is not None:
            return json.dumps(json.loads(message.function_call.arguments))
        else:
            return message.content

    @staticmethod
    def _is_chat():
        return True

    def _get_model_names(self):
        return [combo["model"] for combo in self.argument_combos]

    def _get_prompts(self):
        return [self.prompt_keys[str(combo["messages"][-1]["content"])] for combo in self.argument_combos]

    def _get_state(self):
        partial_col_names = self.partial_df.columns.tolist()
        score_col_names = self.score_df.columns.tolist()
        state_params = {
            "prompt_keys": self.prompt_keys,
            "all_args": self.all_args,
            "partial_col_names": partial_col_names,
            "score_col_names": score_col_names,
        }
        state = (
            state_params,
            self.full_df,
        )
        print("Creating state of experiment...")
        return state

    def save_experiment(self, name: Optional[str] = None):
        r"""
        name (str, optional): Name of the experiment. This is optional if you have previously loaded an experiment
            into this object.
        """
        if name is None and self._experiment_id is None:
            raise RuntimeError("Please provide a name for your experiment.")
        if self.full_df is None:
            raise RuntimeError("Cannot save empty experiment. Please run it first.")
        if os.environ["HEGELAI_API_KEY"] is None:
            raise PermissionError("Please set HEGELAI_API_KEY (e.g. os.environ['HEGELAI_API_KEY']).")
        state = self._get_state()
        url = f"{HEGEL_BACKEND_URL}/sdk/save"
        headers = {
            "Content-Type": "application/octet-stream",  # Use a binary content type for pickled data
            "Authorization": os.environ["HEGELAI_API_KEY"],
        }
        print("Sending HTTP POST request...")
        data = pickle.dumps((name, self._experiment_id, self._experiment_type, state))
        response = requests.post(url, data=data, headers=headers)
        self._experiment_id = response.json().get("experiment_id")
        self._revision_id = response.json().get("revision_id")
        return response

    @classmethod
    def load_experiment(cls, experiment_id: str):
        r"""
        experiment_id (str): experiment ID of the experiment that you wish to load.
        """
        if os.environ["HEGELAI_API_KEY"] is None:
            raise PermissionError("Please set HEGELAI_API_KEY (e.g. os.environ['HEGELAI_API_KEY']).")

        url = f"{HEGEL_BACKEND_URL}/sdk/get/experiment/{experiment_id}"
        headers = {
            "Content-Type": "application/octet-stream",  # Use a binary content type for pickled data
            "Authorization": os.environ["HEGELAI_API_KEY"],
        }
        print("Sending HTTP GET request...")
        response = requests.get(url, headers=headers)
        if response.status_code == 200:  # Note that state should not have `name` included
            new_experiment_id, revision_id, experiment_type_str, state = pickle.loads(response.content)
            if new_experiment_id != experiment_id:
                raise RuntimeError("Experiment ID mismatch between request and response.")
            return cls._load_state(state, experiment_id, revision_id, experiment_type_str)
        else:
            print(f"Error: {response.status_code}, {response.text}")

    @classmethod
    def load_revision(cls, revision_id: str):
        r"""
        revision_id (str): revision ID of the experiment that you wish to load.
        """
        if os.environ["HEGELAI_API_KEY"] is None:
            raise PermissionError("Please set HEGELAI_API_KEY (e.g. os.environ['HEGELAI_API_KEY']).")

        url = f"{HEGEL_BACKEND_URL}/sdk/get/revision/{revision_id}"
        headers = {
            "Content-Type": "application/octet-stream",  # Use a binary content type for pickled data
            "Authorization": os.environ["HEGELAI_API_KEY"],
        }
        print("Sending HTTP GET request...")
        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            experiment_id, new_revision_id, experiment_type_str, state = pickle.loads(response.content)
            if new_revision_id != revision_id:
                raise RuntimeError("Revision ID mismatch between request and response.")
            return cls._load_state(state, experiment_id, revision_id, experiment_type_str)
        else:
            print(f"Error: {response.status_code}, {response.text}")

    @classmethod
    def _load_state(cls, state, experiment_id: str, revision_id: str, experiment_type_str: str):
        (
            state_params,
            full_df,
        ) = state
        if experiment_type_str != cls._experiment_type:
            raise RuntimeError(
                f"The Experiment Type you are trying to load is {experiment_type_str},"
                "which does not match the current class."
            )

        all_args, prompt_keys = state_params["all_args"], state_params["prompt_keys"]
        experiment = cls(all_args["model"], all_args["messages"])
        experiment.prompt_keys = prompt_keys
        experiment.all_args = all_args
        experiment.full_df = pd.DataFrame(full_df)
        experiment.partial_df = (
            experiment.full_df[state_params["partial_col_names"]].copy() if experiment.full_df is not None else None
        )
        experiment.score_df = (
            experiment.full_df[state_params["score_col_names"]].copy() if experiment.full_df is not None else None
        )
        experiment._experiment_id = experiment_id
        experiment._revision_id = revision_id
        print("Loaded experiment.")
        return experiment

    def _validate_arg_key(self, arg_name: str) -> None:
        import inspect

        signature = inspect.signature(self.__init__)
        name_exceptions = {"azure_openai_service_configs"}

        if arg_name in [param.name for param in signature.parameters.values()] and arg_name not in name_exceptions:
            return
        else:
            raise RuntimeError("Provided argument name does not match known argument names.")

    def run_partial(self, **kwargs):
        r"""
        Run experiment with against one parameter, which can be existing or new. The new result will
        be appended to any existing DataFrames.

        If the argument value did not exist before, it will be added to the list of argument combinations
        that will be executed in the next run.

        e.g. `experiement.run_partial({model: 'gpt-4'})`
        """
        print("Running partial experiment...")
        if len(kwargs) > 1:
            raise RuntimeError("Not supported.")
        arg_name, arg_value = list(kwargs.items())[0]

        orginal_arg_value = arg_value
        if arg_name == "messages" and isinstance(arg_value, PromptSelector):
            arg_value = arg_value.for_openai_chat()

        partial_all_args = copy.deepcopy(self.all_args)
        partial_all_args[arg_name] = [arg_value]

        partial_argument_combos = [
            dict(zip(partial_all_args, val)) for val in itertools.product(*partial_all_args.values())
        ]
        original_n_results = len(self.queue.get_results()) if self.queue else 0

        # Execute partial experiment
        for combo in partial_argument_combos:
            self.queue.enqueue(
                self.completion_fn,
                # We need to filter out defaults that are invalid JSON from the request
                {k: v for k, v in combo.items() if (v is not None) and (v != float("inf"))},
            )

        # Verify new results are added
        if len(self.queue.get_results()) - original_n_results == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException

        # Currently, it always append new rows to the results.
        # In the future, we may want to replace existing rows instead.
        self._construct_result_dfs(self.queue.get_input_args(), self.queue.get_results(), self.queue.get_latencies())

        # If `arg_value` didn't exist before, add to `argument_combos`, which will be used in the next `.run()`
        if arg_value not in self.all_args[arg_name]:
            if arg_name == "messages":
                if isinstance(orginal_arg_value, PromptSelector):
                    self.prompt_keys[
                        str(orginal_arg_value.for_openai_chat()[-1]["content"])
                    ] = orginal_arg_value.for_llama()
                else:
                    self.prompt_keys.append(arg_value)
            self.all_args[arg_name].append(arg_value)
            self.prepare()

    def run_one(
        self,
        model: str,
        messages: Union[List[Dict[str, str]], PromptSelector],
        temperature: Optional[float] = 1.0,
        top_p: Optional[float] = 1.0,
        n: Optional[int] = 1,
        stream: Optional[bool] = False,
        stop: Optional[List[str]] = None,
        max_tokens: Optional[int] = float("inf"),
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        logit_bias: Optional[Dict] = None,
        response_format: Optional[dict] = None,
        seed: Optional[int] = None,
        functions: Optional[Dict] = None,
        function_call: Optional[Dict[str, str]] = None,
    ):
        r"""
        Execute one particular configuration of the experiment and add that to the result DataFrame.

        Unlike `run_partial`, this doesn't change the argument combination of the experiment.
        """
        kwargs = {
            "model": model,
            "messages": messages,
            "temperature": temperature,
            "top_p": top_p,
            "n": n,
            "stream": stream,
            "stop": stop,
            "max_tokens": max_tokens,
            "presence_penalty": presence_penalty,
            "frequency_penalty": frequency_penalty,
            "logit_bias": logit_bias,
            "response_format": response_format,
            "seed": seed,
            "functions": functions,
            "function_call": function_call,
        }
        kwargs = {k: v for k, v in kwargs.items() if (v is not None) and (v != float("inf"))}

        original_n_results = len(self.queue.get_results()) if self.queue else 0
        self.queue.enqueue(
            self.completion_fn,
            kwargs,
        )
        if len(self.queue.get_results()) - original_n_results != 1:
            print(original_n_results)
            print(len(self.queue.get_results()))
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException

        self._construct_result_dfs(self.queue.get_input_args(), self.queue.get_results(), self.queue.get_latencies())

    def get_table(self, get_all_cols: bool = False) -> pd.DataFrame:
        columns_to_hide = [
            "stream",
            "response_id",
            "response_choices",
            "response_created",
            "response_created",
            "response_object",
            "response_model",
            "response_system_fingerprint",
            "revision_id",
            "log_id",
        ]

        if get_all_cols:
            return self.full_df
        else:
            table = self.full_df
            columns_to_hide.extend(
                [
                    col
                    for col in ["temperature", "top_p", "n", "presence_penalty", "frequency_penalty"]
                    if col not in self.partial_df.columns
                ]
            )
            for col in columns_to_hide:
                if col in table.columns:
                    table = table.drop(col, axis=1)
            return table

    # def _update_values_in_dataframe(self):
    #     r"""
    #     If, in the future, we wish to update existing values rather than appending to the end of the row.
    #
    #     # Consider doing a merge left here
    #     #       1. Identify what input_args columns exist
    #     #       2. Use those columns names for pandas to do a merge left
    #     #       3. If a value (from evals mostly) doesn't exist in the new one, put as NaN or empty
    #     #       4. If 1 has the key combo but 2 doesn't, mkae sure to keep the one from 1
    #     #       5. Make sure `scores_df` is correct
    #     # Alternatively, find the index and overwrite those DataFrame rows, where each row is a `pd.Series`.
    #     """
    #     pass



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List, Optional, Union
from operator import itemgetter
import base64
from collections import defaultdict
import itertools
import logging
from IPython import display
from tabulate import tabulate
import pandas as pd
import sentry_sdk
import os

try:
    import cv2
except ImportError:
    cv2 = None

try:
    import pymongo
except ImportError:
    pymongo = None

from prompttools.requests.request_queue import RequestQueue

# from ..widgets.feedback import FeedbackWidgetProvider
# from ..widgets.comparison import ComparisonWidgetProvider
from ..widgets.utility import is_interactive
from .error import PromptExperimentException
from ._utils import _get_dynamic_columns

pd.set_option("display.max_colwidth", 0)


class Experiment:
    r"""
    Base class for experiment. This should not be used directly, please use the subclasses instead.
    """

    completion_fn: Callable
    all_args: Dict

    def __init__(self):
        self.queue = RequestQueue()
        self.argument_combos: list[dict] = []
        self.full_df, self.partial_df, self.score_df = None, None, None
        self.image_experiment = False
        self._experiment_id = None
        self._revision_id = None
        try:
            if "SENTRY_OPT_OUT" not in os.environ:
                sentry_sdk.capture_message(f"Initializing {self.__class__.__name__}", "info")
        except Exception:
            pass
        # self.feedback_widget_provider = FeedbackWidgetProvider(
        #     self.completion_fn, self._aggregate_metric, self._get_human_eval_listener
        # )
        # self.comparison_widget_provider = ComparisonWidgetProvider(
        #     self.completion_fn,
        #     self._aggregate_comparison,
        #     self._get_comparison_listener,
        # )

    @classmethod
    def initialize(cls, test_parameters: dict[str, list], frozen_parameters: dict):
        r"""
        An alternate way to initialize an experiment by specifying which parameters should be tested
        and which ones should be frozen. If a parameter is not specified, the default value (if exists)
        for the parameter will be used.

        This allows you to easily initialize an experiment **without** wrapping every parameter in a list.

        Note:
            - For a given experiment, some parameters must be specified (e.g. the ``model`` parameter
              for OpenAI Chat Experiment). See the experiment's ``__init__`` method.
            - Each of ``test_parameters``'s values should be a ``list``, but not for ``frozen_parameters``.

        Args:
            test_parameters (dict[str, list]): parameters that are being tested. A list of multiple test values
                should be the value (e.g. ``{model: ["gpt-3.5-turbo", "gpt-4"], temperature: [0,0. 1.0]}``)
            frozen_parameters (dict): parameters that are intended to be frozen across different configuration.
                There is no need to wrap the value in a list. (e.g. ``{top_p: 1.0, presence_penalty: 0.0}``)

        Example:
            >>> from prompttools.experiment import OpenAIChatExperiment
            >>> test_parameters = {"model": ["gpt-3.5-turbo", "gpt-4"]}
            >>> messages = [{"role": "user", "content": "Who was the first president?"}]
            >>> frozen_parameters = {"top_p": 1.0, "messages": messages}
            >>> experiment = OpenAIChatExperiment.initialize(test_parameters, frozen_parameters)
        """
        frozen_parameters = {k: [v] for k, v in frozen_parameters.items()}
        return cls(**test_parameters, **frozen_parameters)

    @staticmethod
    def _is_chat():
        return False

    # def _get_human_eval_listener(self, i: int) -> Callable:
    #     def listener(change):
    #         self.score_df["feedback"][i] = change["new"]
    #
    #     return listener

    # def _get_comparison_listener(self, index: int) -> Callable:
    #     def listener(change):
    #         new_index = self.comparison_index_translation(index)
    #         self.score_df["comparison"][new_index] = change["new"]
    #
    #     return listener

    # def _aggregate_comparison(
    #     self,
    #     table: pd.DataFrame,
    #     agg_column: int = 0,
    #     is_average: bool = False,
    # ) -> Dict[str, int]:
    #     # TODO: This could be a group by
    #     prompt_scores = defaultdict(int)
    #     prompt_counts = defaultdict(int)
    #     for index, row in enumerate(table.iterrows()):
    #         key = str(row[agg_column])
    #         new_index = self.comparison_index_translation(index)
    #         prompt_scores[key] += self.score_df["comparison"][new_index]
    #         prompt_counts[key] += 1
    #     if is_average:
    #         for k, v in prompt_scores.items():
    #             prompt_scores[k] = v / prompt_counts[k]
    #     sorted_scores = dict(sorted(prompt_scores.items(), key=lambda item: item[1], reverse=True))
    #     return sorted_scores

    def _aggregate_metric(
        self,
        table: pd.DataFrame,
        metric_name: str,
        agg_column: str,
        is_average: bool = False,
    ) -> Dict[str, int]:
        # TODO: This could be a group by

        prompt_scores = defaultdict(int)
        prompt_counts = defaultdict(int)
        for index, row in table.iterrows():
            key = str(row[agg_column])
            prompt_scores[key] += self.score_df[metric_name][index]
            prompt_counts[key] += 1
        if is_average:
            for k, v in prompt_scores.items():
                prompt_scores[k] = v / prompt_counts[k]
        sorted_scores = dict(sorted(prompt_scores.items(), key=lambda item: item[1], reverse=True))
        return sorted_scores

    def prepare(self) -> None:
        r"""
        Creates argument combinations by taking the cartesian product of all inputs.
        """
        self.argument_combos = [dict(zip(self.all_args, val)) for val in itertools.product(*self.all_args.values())]

    def run(
        self,
        runs: int = 1,
        clear_previous_results: bool = False,
    ) -> None:
        r"""
        Create tuples of input and output for every possible combination of arguments.

        Note:
            If you overwrite this method in a subclass, make sure your method calls ``_construct_result_dfs``
            in order to save the results from your run as DataFrames. Then, they can later be used
            for evaluation, aggregation, and persistence.

        Args:
            runs (int): number of times to execute each possible combination of arguments, defaults to 1.
            clear_previous_results (bool): clear previous results before running
        """
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        if clear_previous_results:
            self.queue = RequestQueue()
            self.full_df, self.partial_df, self.score_df = None, None, None
        original_n_results = len(self.queue.get_results()) if self.queue else 0
        for combo in self.argument_combos:
            for _ in range(runs):
                self.queue.enqueue(
                    self.completion_fn,
                    # We need to filter out defaults that are invalid JSON from the request
                    {k: v for k, v in combo.items() if (v is not None) and (v != float("inf"))},
                )
        number_of_new_results = len(self.queue.get_results()) - original_n_results
        if number_of_new_results == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException
        results = self.queue.get_results()[-number_of_new_results:]
        input_args = self.queue.get_input_args()[-number_of_new_results:]
        latencies = self.queue.get_latencies()[-number_of_new_results:]
        self._construct_result_dfs(input_args, results, latencies)

    def _construct_result_dfs(
        self,
        input_args: list[dict[str, object]],
        results: list[dict[str, object]],
        latencies: list[float],
        extract_response_equal_full_result: bool = False,
        response_extractors: Optional[dict[str, Callable]] = None,
    ):
        r"""
        Takes in the input, results, and other metrics from the experiment's run, and construct a few DataFrames that
        contain all relevant data (i.e. input arguments, results, evaluation metrics).

        These DataFrames can later be used for evaluation, aggregation, or storing them for persistence.

        Note:
            - If your subclass of ``Experiment`` has a custom ``run`` method. You should consider overwriting this
              method. In particular, you likely would want to define how to extract the response from LLM's result
              and save that into ``response_df`` below. ChromaDBExperiment provides an example of this.
            - The inputs should all share the same length.

        Args:
             input_args (list[dict[str, object]]): list of dictionaries, where each of them is a set of
                 input argument that was passed into the model
             results (list[dict[str, object]]): list of responses from the model
             latencies (list[float]): list of latency measurements
             extract_response_equal_full_result (bool): if ``True``, ``result_df`` will only contain
                 the extracted response, lead to a simpler (but incomplete) columns of results.
             response_extractors (Optional[dict[str, Callable]]): Optional dictionary of response extractors,
                 defaults to ``None``, which will use the ``_extract_responses`` defined by the class.
                 The key of the dictionary will be the name of the resulting column, the value of the dictionary
                 will be an extractor function that accepts the response from the model and returns a value.
        """
        # `input_arg_df` contains all all input args
        input_arg_df = pd.DataFrame(input_args)
        # `dynamic_input_arg_df` contains input args that has more than one unique values
        dynamic_input_arg_df = _get_dynamic_columns(input_arg_df)

        # `response_df` contains the extracted response (often being the text response)
        if response_extractors is None:
            response_df = pd.DataFrame({"response": [self._extract_responses(result) for result in results]})
        else:
            res_dict = {}
            for col_name, extractor in response_extractors.items():
                res_dict[col_name] = [extractor(result) for result in results]
            response_df = pd.DataFrame(res_dict)
        # `result_df` contains everything returned by the completion function
        if extract_response_equal_full_result:
            result_df = response_df
        else:
            # Handle the case where `input_arg_df` has the same column names as `result_df`
            try:
                results = [r.model_dump() for r in results]  # For turing OpenAI response in to dict
            except Exception:
                pass
            result_df = pd.DataFrame(results)
            common_columns = set(input_arg_df.columns) & set(result_df.columns)
            result_df = result_df.add_prefix("response_") if common_columns else result_df
            result_df = pd.concat([response_df, result_df], axis=1)

        # `score_df` contains computed metrics (e.g. latency, evaluation metrics)
        new_score_df = pd.DataFrame({"latency": latencies})
        self.score_df = new_score_df if self.score_df is None else pd.concat([self.score_df, new_score_df])

        # `partial_df` contains some input arguments, extracted responses, and score
        new_partial_df = pd.concat([dynamic_input_arg_df, response_df, new_score_df], axis=1)
        self.partial_df = new_partial_df if self.partial_df is None else pd.concat([self.partial_df, new_partial_df])

        # `full_df` contains all input arguments, responses, and score
        new_full_df = pd.concat([input_arg_df, result_df, new_score_df], axis=1)
        self.full_df = new_full_df if self.full_df is None else pd.concat([self.full_df, new_full_df])

    def get_table(self, get_all_cols: bool = False) -> pd.DataFrame:
        r"""
        Get the DataFrame in one of two versions:
        1. ``get_all_cols = False`` - good for visualization. This contains dynamic (non-frozen) input arguments,
            the text response, and scores (e.g. latency and metrics generated from evaluation).
        2. ``get_all_cols = True`` - good for full result. This contains full data with all
            input arguments (including frozen ones), full model response (not just the text response), and scores.

        Args:
            get_all_cols (bool): defaults to ``False``. If ``True``, it will return the full data with all
                input arguments (including frozen ones), full model response (not just the text response), and scores.
        """
        if self.full_df is None:
            logging.info("Running first...")
            self.run()
        if get_all_cols:
            return self.full_df
        else:
            return self.partial_df

    def visualize(self, get_all_cols: bool = False, pivot: bool = False, pivot_columns: list = []) -> None:
        r"""
        Visualize the DataFrame in one of two versions:
        1. ``get_all_cols = False`` - good for visualization. This contains dynamic (non-frozen) input arguments,
            the text response, and scores (e.g. latency and metrics generated from evaluation).
        2. ``get_all_cols = True`` - good for full result. This contains full data with all
            input arguments (including frozen ones), full model response (not just the text response), and scores.

        Args:
            get_all_cols (bool): defaults to ``False``. If ``True``, it will visualize the full data with all
                input arguments (including frozen ones), full model response (not just the text response), and scores.
        """
        if pivot:
            table = self.pivot_table(pivot_columns, get_all_cols=get_all_cols)
        else:
            table = self.get_table(get_all_cols)
        if is_interactive() and self.image_experiment:
            # revert to color, display as image in notebook cells
            # for user friendly experience
            table["response"] = table["response"].map(lambda x: cv2.cvtColor(x, cv2.COLOR_GRAY2BGR))
            table["response"] = table["response"].map(lambda x: self.cv2_image_to_base64(x))
            table["response"] = table["response"].apply(self.display_image_html)
            display.display(display.HTML(table.to_html(escape=False)))
        elif is_interactive():
            display.display(table)
        else:
            logging.getLogger().setLevel(logging.INFO)
            logging.info(tabulate(table, headers="keys", tablefmt="psql"))

    def cv2_image_to_base64(self, image):
        if cv2 is None:
            raise ModuleNotFoundError(
                "Package `cv2` is required to be installed to use this experiment."
                "Please use `pip opencv-python` to install the package"
            )
        _, buffer = cv2.imencode(".png", image)
        return base64.b64encode(buffer).decode("utf-8")

    def display_image_html(self, base64_string):
        return f'<img src="data:image/png;base64,{base64_string}" width="150" height="150"/>'

    def evaluate(
        self,
        metric_name: str,
        eval_fn: Callable,
        static_eval_fn_kwargs: dict = {},
        image_experiment: bool = False,
        **eval_fn_kwargs,
    ) -> None:
        """
        Using the given evaluation function that accepts a row of data, compute a new column with the evaluation
        result. Each row of data generally contain inputs, model response, and other previously computed metrics.

        Args:
            metric_name (str): name of the metric being computed
            eval_fn (Callable): an evaluation function that takes in a row from pd.DataFrame
                and optional keyword arguments
            static_eval_fn_kwargs (dict): keyword args for ``eval_fn`` that are consistent for all rows
            eval_fn_kwargs (Optional[list]): keyword args for ``eval_fn`` that may be different for each row.
                Each value entered here should be a list, and the length of the list should be
                the same as the number of responses in the experiment's result. The ``i``th element of the list will be
                passed to the evaluation function to evaluate the ``i``th row.

        Example:
            >>> from prompttools.utils import validate_json_response
            >>> experiment.evaluate("is_json", validate_json_response,
            >>>                     static_eval_fn_kwargs={"response_column_name": "response"})
        """
        self.image_experiment = image_experiment
        if metric_name in self.score_df.columns:
            logging.warning(metric_name + " is already present, skipping.")
            return
        res = []
        table = self.get_table(get_all_cols=True)
        for i, row in table.iterrows():
            curr_kwargs = static_eval_fn_kwargs.copy()
            for k, v in eval_fn_kwargs.items():
                curr_kwargs[k] = v[i]
            res.append(eval_fn(row, **curr_kwargs))
        self._update_score(metric_name, res)

    def _update_score(self, metric_name: str, res) -> None:
        self.score_df[metric_name] = res
        self.partial_df[metric_name] = res
        self.full_df[metric_name] = res

    def pivot_table(
        self, pivot_columns: List[str], response_value_name: Optional[str] = None, get_all_cols: bool = False
    ) -> pd.DataFrame:
        """
        Returns a pivoted DataFrame.

        Args:
            pivot_columns (List[str]): two column names (first for pivot row, second for pivot column)
                that serve as indices the pivot table
            response_value_name (Optional[str]): name of the column to aggregate.
            get_all_cols (bool): defaults to ``False``. If ``True``, it will visualize the full data with all
                input arguments (including frozen ones), full model response (not just the text response), and scores.
        """
        df = self.get_table(get_all_cols)
        pivot_df = pd.pivot_table(
            df,
            values=response_value_name,
            index=[pivot_columns[1]],
            columns=[pivot_columns[0]],
            aggfunc=lambda x: x.iloc[0],
        )
        return pivot_df

    # def gather_feedback(self, pivot_data: Dict[str, object], pivot_columns: List[str]) -> None:
    #     """
    #     This method creates a table to gather human feedback from a notebook interface.
    #
    #     Args:
    #         pivot_data (Dict[str, object]): dictionary that contains additional data or metadata related to the input
    #         pivot_columns (List[str]): two column names (first for pivot row, second for pivot column)
    #             that serve as indices the pivot table
    #     """
    #     if not is_interactive():
    #         logging.warning("This method only works in notebooks.")
    #         return
    #     table = self.get_table(get_all_cols=True)
    #     self.score_df["feedback"] = [1] * len(table.index))
    #     self.feedback_widget_provider.set_pivot_columns(pivot_columns)
    #     items = self.feedback_widget_provider.get_header_widgets()
    #     for row in table.iterrows():
    #         items += self.feedback_widget_provider.get_row_widgets(*row)
    #     items += self.feedback_widget_provider.get_footer_widgets(table)
    #     self.feedback_widget_provider.display(items)

    # def compare(self, primary_model: str, pivot_columns: List[str]) -> None:
    #     """
    #     This method creates a table to gather human feedback from a notebook interface.
    #     """
    #     if not is_interactive():
    #         logging.warning("This method only works in notebooks.")
    #         return
    #     table = self.get_table(pivot_data={}, pivot_columns=pivot_columns, pivot=True)
    #     self.score_df["comparison"] = [1] * len(table.index))
    #     self.comparison_index_translation = lambda i: i * len(table.columns)
    #     self.comparison_widget_provider.set_models(table.columns)
    #     items = self.comparison_widget_provider.get_header_widgets()
    #     for index, row in enumerate(table.iterrows()):
    #         items += self.comparison_widget_provider.get_row_widgets(index, row[1])
    #     items += self.comparison_widget_provider.get_footer_widgets(table)
    #     self.comparison_widget_provider.display(items)
    def aggregate(self, metric_name, column_name, is_average=False):
        r"""
        Aggregates a metric for a given column and displays to the user.

         Args:
            metric_name (str): metric to aggregate
            column_name (str): column to base the aggregation on
            is_average (bool): if ``True``, compute the average for the metric, else compute the total
        """
        if self.score_df is None or metric_name not in self.score_df.columns:
            logging.warning("Can't find " + metric_name + " in scores. Did you run `evaluate`?")
            return
        table = self.get_table(get_all_cols=False)
        sorted_scores = self._aggregate_metric(table, metric_name, column_name, is_average)
        if is_interactive():
            import matplotlib.pyplot as plt
            import os

            # Import style file, assumes same dir as experiment.py
            style_path = os.path.join(os.path.dirname(__file__), "style.mplstyle")
            plt.style.use(style_path)

            # Define the custom colors
            custom_colors = [
                "black",
                "#7e1e9c",
                "#15b01a",
                "#448ee4",
                "#ff7fa7",
                "#029386",
            ]

            plt.ylabel("Latency (s)")

            # Cycle through the custom colors when creating the bars
            for i, (label, value) in enumerate(sorted_scores.items()):
                plt.bar(i, value, align="center", color=custom_colors[i % len(custom_colors)])

            plt.xticks(range(len(sorted_scores)), list(sorted_scores.keys()))
            plt.show()

    def rank(
        self,
        metric_name: str,
        is_average: bool,
        agg_column: str,
        get_all_cols: bool = False,
    ) -> Dict[str, int]:
        """
        Using pivot data, groups the data by the first pivot column to
        get scores, and sorts descending. For example, using pivot data of
        (prompt_template, user_input), a metric of latency, and is_average=True,
        we rank prompt templates by their average latency in the test set.

        Args:
            metric_name (str): metric to aggregate over
            is_average (bool): if ``True``, compute the average for the metric, else compute the total
            agg_column (str): column to aggregate over
            get_all_cols (bool): defaults to ``False``. If ``True``, it will return the full data with all
                input arguments (including frozen ones), full model response (not just the text response), and scores.
        """
        if self.score_df is None or metric_name not in self.score_df.columns:
            logging.warning("Can't find " + metric_name + " in scores. Did you run `evaluate`?")
            return
        table = self.get_table(get_all_cols=get_all_cols)
        sorted_scores = self._aggregate_metric(table, metric_name, agg_column, is_average)
        return sorted_scores

    @staticmethod
    def _extract_responses(output: Dict[str, object]) -> list[str]:
        raise NotImplementedError("This should be implemented by a subclass of `Experiment`.")

    def to_csv(
        self,
        path: str,
        get_all_cols: bool = True,
        **kwargs,
    ):
        r"""
        Export the results to a CSV file. If the experiment has not been executed, it will run.

        Args:
            path (str): path/buffer to write the CSV output
            get_all_cols (bool): defaults to ``False``. If ``True``, it will return the full data with all
                input arguments (including frozen ones), full model response (not just the text response), and scores.
            **kwargs: optional arguments passed to ``pd.DataFrame.to_csv()``
        """
        table = self.get_table(get_all_cols=get_all_cols)
        table.to_csv(path, **kwargs)

    def to_pandas_df(self, get_all_cols: bool = True, from_streamlit: bool = False):
        r"""
        Return the results as a ``pandas.DataFrame``. If the experiment has not been executed, it will run.

        Args:
            get_all_cols (bool): defaults to ``False``. If ``True``, it will return the full data with all
                input arguments (including frozen ones), full model response (not just the text response), and scores.
        """
        if from_streamlit:
            self.run()
        return self.get_table(get_all_cols=get_all_cols)

    def to_json(
        self,
        path: Optional[str] = None,
        get_all_cols: bool = True,
        **kwargs,
    ):
        r"""
        Export the results to a JSON file. If the experiment has not been executed, it will run.

        Args:
            path (Optional[str]): path/buffer to write the JSON output, defaults to ``None`` which returns
                the JSON as a `dict`
            get_all_cols (bool): defaults to ``False``. If ``True``, it will return the full data with all
                input arguments (including frozen ones), full model response (not just the text response), and scores.
            **kwargs: optional arguments passed to ``pd.DataFrame.to_json()``
        """
        table = self.get_table(get_all_cols=get_all_cols)
        if path is None:
            return table.to_json(**kwargs)
        else:
            return table.to_json(path, **kwargs)

    def to_lora_json(
        self,
        instruction_extract: Union[str, Callable],
        input_extract: Union[str, Callable],
        output_extract: Union[str, Callable],
        path: Optional[str] = None,
        **kwargs,
    ):
        r"""
        Export the results to a LoRA-format JSON file for fine-tuning.
        If the experiment has not been executed, it will run.

        Args:
            instruction_extract (Union[str, Callable]): column name, or an extractor function that will accept a row
                of the result table and return a value assigned to ``"instruction"`` entry in the JSON file
            input_extract (Union[str, Callable]): column name, or an extractor function that will accept a row
                of the result table and return a value assigned to ``"input"`` entry in the JSON file
            output_extract (Union[str, Callable]): column name, or an extractor function that will accept a row
                of the result table and return a value assigned to ``"output"`` entry in the JSON file
            path (Optional[str]): path/buffer to write the JSON output, defaults to ``None`` which returns
                the JSON as a `dict`
            **kwargs: optional arguments passed to ``pd.DataFrame.to_json()``
        """
        if isinstance(instruction_extract, str):
            instruction_extract = itemgetter(instruction_extract)
        if isinstance(input_extract, str):
            input_extract = itemgetter(input_extract)
        if isinstance(output_extract, str):
            output_extract = itemgetter(output_extract)
        df = self.to_pandas_df(get_all_cols=True)
        extracted_data = df.apply(
            lambda row: {
                "instruction": instruction_extract(row),
                "input": input_extract(row),
                "output": output_extract(row),
            },
            axis=1,
        )
        if "orient" not in kwargs:
            kwargs["orient"] = "records"
        if "indent" not in kwargs:
            kwargs["indent"] = 2

        if path:
            extracted_data.to_json(path, **kwargs)
        else:
            return extracted_data.to_json(**kwargs)

    # TODO: Add MongoDB local instruction (maybe include docker)
    def to_mongo_db(self, mongo_uri: str, database_name: str, collection_name: str) -> None:
        r"""
        Insert the results of the experiment into MongoDB for persistence.

        Note:
            - You need to install the ``pymongo`` package to use this method.
            - You need to run a local or remote instance of MongoDB in order to store the data.

        Args:
            mongo_uri (str): a connection string to the target MongoDB
            database_name (str): name of the MongoDB database
            collection_name (str): name of the MongoDB collection
        """
        if pymongo is None:
            raise ModuleNotFoundError(
                "Package `pymongo` is required to be installed to use this method."
                "Please use `pip install pymongo` to install the package"
            )
        if self.full_df is None:
            logging.info("Running first...")
            self.run()
        client = pymongo.MongoClient(mongo_uri)
        db = client[database_name]
        collection = db[collection_name]
        collection.insert_many(self.full_df.to_dict("records"))
        logging.info(f"Inserted results in {database_name}'s collection {collection_name}.")
        client.close()

    def to_markdown(self):
        markdown = self.to_pandas_df().to_markdown()
        return markdown

    def _get_model_names(self):
        pass

    def _get_prompts(self):
        pass

    def _get_state(self):
        raise NotImplementedError("Should be implemented by specific harness class.")

    @classmethod
    def _load_state(cls, state, experiment_id: str, revision_id: str, experiment_type_str: str):
        raise NotImplementedError("Should be implemented by specific harness class.")



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


from typing import Callable, Dict, List, Optional
import pandas.core.series
import json
import re

KEY_EXTRACTION_REGEX = r'"([^"]+?)"\s*:'


def strip_outer_brackets(text: str) -> str:
    r"""
    Removes all chars outside the first '{' and the last '}'. Intended to be a pre-processing
    step prior to parsing a string as JSON.

    Args:
        text(str): the text to process
    """
    first_brace = text.find("{")
    last_brace = text.rfind("}")
    return text[first_brace : last_brace + 1]


def sample_pre_process_fn(text: str):
    r"""
    An example pre-processing that you may use before attempting to parse a string as JSON.
    This function removes all chars outside the first '{' and the last '}'. Then,
    it removes ``"\\n"``.

    This function should be modified depending on your LLM's output.

    Args:
        text(str): the text to process
    """
    text = strip_outer_brackets(text)
    text = text.replace("\\n", "")
    return text


def validate(text: str, pre_process_fn: Optional[Callable] = None):
    r"""
    Validates that the generated text is JSON.

    Args:
        text (str): The generated text, which should be valid JSON.
        pre_process_fn (Callable[str, str]): a function to pre-process the text response from the LLM before attempting
            to parse the string as JSON. Look at ``validate_json.sample_pre_process_fn`` as an example.
    """
    if pre_process_fn:
        text = pre_process_fn(text)
    try:
        json.loads(text)
    except ValueError:
        return 0.0
    return 1.0


def validate_keys(text: str, valid_keys: List[str]):
    r"""
    Guarantees that all keys in the generated JSON are valid.

    Args:
        text (str): The generated text, which should be valid JSON.
        valid_keys (List[str]): A list of valid keys which may appear in the JSON.
    """
    match = re.search(text, KEY_EXTRACTION_REGEX)
    for group in match.groups():
        if group not in valid_keys:
            return 0.0
    return 1.0


def validate_json_response(row: pandas.core.series.Series, response_column_name: str = "response") -> float:
    r"""
    Validate whether ``response`` string is in a valid JSON format.

    Args:
        row (pandas.core.series.Series): A row of data from the full DataFrame (including input, model response, other
            metrics, etc).
        response_column_name (str): name of the column that contains the model's response, defaults to ``"response"``
    """
    return validate(row[response_column_name])


def evaluate(prompt: str, response: str, metadata: Dict) -> float:
    r"""
    Validate whether ``response`` string is in a valid JSON format.

    Args:
        prompt (str): Not used.
        response (str): the string that will be validated
        metadata (dict): Not used.
    """
    return validate(response)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Type
import jinja2
import pandas as pd

from .harness import ExperimentationHarness, Experiment
from typing import Optional
from copy import deepcopy
from .utility import is_interactive
from IPython import display
from tabulate import tabulate
import logging


def _render_messages_openai_chat(message_template: list[dict], user_input: dict, environment):
    rendered_message = deepcopy(message_template)
    sys_msg_template = environment.from_string(rendered_message[0]["content"])
    user_msg_template = environment.from_string(rendered_message[-1]["content"])
    rendered_message[0]["content"] = sys_msg_template.render(**user_input)
    rendered_message[-1]["content"] = user_msg_template.render(**user_input)
    return rendered_message


class ChatPromptTemplateExperimentationHarness(ExperimentationHarness):
    r"""
    An experimentation harness used to test various prompt templates for chat models.
    We use `jinja` templates, e.g. "Answer the following question: {{input}}".

    Args:
        experiment (Type[Experiment]): The experiment constructor that you would like to execute within the harness
            (e.g. ``prompttools.experiment.OpenAICompletionExperiment``)
        model_name (str): The name of the model.
        message_templates (List[str]): A list of prompt ``jinja``-styled templates. Each template should have two
            messages inside (first system prompt and second a user message).
        user_inputs (List[Dict[str, str]]): A list of dictionaries representing user inputs.
        model_arguments (Optional[Dict[str, object]], optional): Additional arguments for the model.
            Defaults to ``None``. Note that the values are not lists.
    """

    _experiment_type = "Template"
    PIVOT_COLUMNS = ["prompt_template", "user_input"]

    def __init__(
        self,
        experiment: Type[Experiment],
        model_name: str,
        message_templates: list[list[dict]],
        user_inputs: list[dict[str, str]],
        model_arguments: Optional[dict[str, object]] = None,
    ):
        self.environment = jinja2.Environment()
        self.experiment_cls_constructor = experiment
        self.model_name = model_name
        self.message_templates = message_templates
        self.user_inputs = user_inputs
        self.model_arguments = {} if model_arguments is None else model_arguments
        super().__init__()

    def prepare(self) -> None:
        r"""
        Creates prompts from templates to use for the experiment, and then initializes and prepares the experiment.
        """
        # self.input_pairs_dict = {}
        rendered_messages = []
        for mt in self.message_templates:
            for user_input in self.user_inputs:
                rendered_messages.append(_render_messages_openai_chat(mt, user_input, self.environment))
        self.experiment = self.experiment_cls_constructor(
            [self.model_name],
            rendered_messages,
            **self._prepare_arguments(self.model_arguments),
        )
        super().prepare()

    def run(self, clear_previous_results: bool = False):
        if not self.experiment:
            self.prepare()
        super().run(clear_previous_results=clear_previous_results)

        # Add user inputs to DataFrame
        if len(self.experiment.full_df) > 0:
            repeat = len(self.experiment.full_df) // len(self.user_inputs)
            user_inputs = deepcopy(self.user_inputs)
            user_inputs_col_name = "user_inputs"
            user_input_df = pd.DataFrame({user_inputs_col_name: user_inputs * repeat})

            if user_inputs_col_name in self.experiment.full_df.columns:
                self.experiment.full_df = self.experiment.full_df.drop(user_inputs_col_name, axis=1)
            self.experiment.full_df.reset_index(drop=True, inplace=True)

            self.experiment.full_df = pd.concat([user_input_df, self.experiment.full_df], axis=1)
            if user_inputs_col_name in self.experiment.partial_df.columns:
                self.experiment.partial_df = self.experiment.partial_df.drop(user_inputs_col_name, axis=1)
            self.experiment.partial_df.reset_index(drop=True, inplace=True)
            self.experiment.partial_df = pd.concat([user_input_df, self.experiment.partial_df], axis=1)

    def get_table(self, get_all_cols: bool = False) -> pd.DataFrame:
        columns_to_hide = [
            "stream",
            "response_id",
            "response_choices",
            "response_created",
            "response_created",
            "response_object",
            "response_model",
            "response_system_fingerprint",
            "revision_id",
            "log_id",
        ]

        if get_all_cols:
            return self.full_df
        else:
            table = self.full_df
            columns_to_hide.extend(
                [
                    col
                    for col in ["temperature", "top_p", "n", "presence_penalty", "frequency_penalty"]
                    if col not in self.partial_df.columns
                ]
            )
            for col in columns_to_hide:
                if col in table.columns:
                    table = table.drop(col, axis=1)
            return table

    def visualize(self, get_all_cols: bool = False):
        table = self.get_table(get_all_cols)
        if is_interactive():
            display.display(table)
        else:
            logging.getLogger().setLevel(logging.INFO)
            logging.info(tabulate(table, headers="keys", tablefmt="psql"))

    def _get_state(self):
        state_params = {
            "experiment_cls_constructor": self.experiment_cls_constructor,
            "model_name": self.model_name,
            "message_templates": self.message_templates,
            "user_inputs": self.user_inputs,
            "model_arguments": self.model_arguments,
            "child_experiment_state": self.experiment._get_state() if self.experiment else None,
        }
        state = (
            state_params,
            self.experiment.full_df,
        )
        print("Creating state of experiment...")
        return state

    @classmethod
    def _load_state(cls, state, experiment_id: str, revision_id: str, experiment_type_str: str):
        (
            state_params,
            full_df,
        ) = state
        if experiment_type_str != cls._experiment_type:
            raise RuntimeError(
                f"The Experiment Type you are trying to load is {experiment_type_str},"
                "which does not match the current class."
            )

        experiment_cls_constructor = state_params["experiment_cls_constructor"]
        model_name = state_params["model_name"]
        message_templates = state_params["message_templates"]
        user_inputs = state_params["user_inputs"]
        model_arguments = state_params["model_arguments"]
        child_experiment_state = state_params["child_experiment_state"]

        harness = cls(experiment_cls_constructor, model_name, message_templates, user_inputs, model_arguments)
        harness.experiment = experiment_cls_constructor._load_state(
            child_experiment_state, None, None, experiment_cls_constructor._experiment_type
        )
        harness._experiment_id = experiment_id
        harness._revision_id = revision_id
        print("Loaded harness.")
        return harness



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional
from .harness import ExperimentationHarness
from prompttools.experiment import OpenAIChatExperiment


class ChatHistoryExperimentationHarness(ExperimentationHarness):
    r"""
    An experimentation harness used for compare multiple chat histories.

    Args:
        model_name (str): The name of the model.
        chat_histories (List[List[Dict[str, str]]]): A list of chat histories that will be fed into the model.
        model_arguments (Optional[Dict[str, object]], optional): Additional arguments for the model.
            Defaults to ``None``.
    """

    def __init__(
        self,
        model_name: str,
        chat_histories: List[List[Dict[str, str]]],
        model_arguments: Optional[Dict[str, object]] = None,
    ):
        self.experiment_cls_constructor = OpenAIChatExperiment
        self.model_name = model_name
        self.chat_histories = chat_histories
        self.model_arguments = {} if model_arguments is None else model_arguments
        super().__init__()

    def prepare(self) -> None:
        r"""
        Initializes and prepares the experiment.
        """
        self.experiment = self.experiment_cls_constructor(
            [self.model_name],
            self.chat_histories,
            **self._prepare_arguments(self.model_arguments),
        )
        super().prepare()

    def run(self):
        if not self.experiment:
            self.prepare()
        super().run()



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import os
import jinja2
import streamlit as st

from prompttools.selector.prompt_selector import PromptSelector
from prompttools.playground.constants import ENVIRONMENT_VARIABLE, EXPERIMENTS


def render_prompts(templates, vars):
    prompts = []
    for template in templates:
        for var_set in vars:
            environment = jinja2.Environment()
            jinja_template = environment.from_string(template)
            prompts.append(jinja_template.render(**var_set))
    return prompts


@st.cache_data
def load_data(
    model_type,
    model,
    instructions,
    user_inputs,
    temperature=0.0,
    top_p=1,
    max_tokens=None,
    frequency_penalty=0.0,
    presence_penalty=0.0,
    api_key=None,
):
    if api_key:
        os.environ[ENVIRONMENT_VARIABLE[model_type]] = api_key
    selectors = [PromptSelector(instruction, user_input) for instruction in instructions for user_input in user_inputs]

    experiment = None
    if model_type == "LlamaCpp Chat":
        call_params = dict(temperature=[temperature])
        experiment = EXPERIMENTS[model_type]([model], selectors, call_params=call_params)
    elif model_type in {"OpenAI Chat", "OpenAI Completion"}:
        experiment = EXPERIMENTS[model_type](
            [model],
            selectors,
            temperature=[temperature],
            top_p=[top_p],
            max_tokens=[max_tokens],
            frequency_penalty=[frequency_penalty],
            presence_penalty=[presence_penalty],
        )
    elif model_type == "HuggingFace Hub":
        experiment = EXPERIMENTS[model_type]([model], selectors, temperature=[temperature])
    elif model_type == "Anthropic":
        experiment = EXPERIMENTS[model_type]([model], selectors, temperature=[temperature])
    elif model_type == "Google PaLM":
        experiment = EXPERIMENTS[model_type]([model], selectors, temperature=[temperature])
    elif model_type == "Replicate":
        input_kwargs = {"prompt": selectors,
                        "temperature": [temperature]}
        model_specific_kwargs = {model: {}}
        experiment = EXPERIMENTS[model_type]([model], input_kwargs, model_specific_kwargs)

    return experiment.to_pandas_df(True, True)


@st.cache_data
def run_multiple(
    model_types,
    models,
    instructions,
    prompts,
    openai_api_key=None,
    anthropic_api_key=None,
    google_api_key=None,
    hf_api_key=None,
    replicate_api_key=None,
):
    import os

    if openai_api_key:
        os.environ["OPENAI_API_KEY"] = openai_api_key
    if anthropic_api_key:
        os.environ["ANTHROPIC_API_KEY"] = anthropic_api_key
    if google_api_key:
        os.environ["GOOGLE_PALM_API_KEY"] = google_api_key
    if hf_api_key:
        os.environ["HUGGINGFACEHUB_API_TOKEN"] = hf_api_key
    if replicate_api_key:
        os.environ["REPLICATE_API_TOKEN"] = replicate_api_key
    dfs = []
    for i in range(len(models)):
        # TODO Support temperature and other parameters
        selectors = []
        if i + 1 in instructions:
            selectors = [PromptSelector(instructions[i + 1], prompt) for prompt in prompts]
            if model_types[i] == "Replicate":
                input_kwargs = {"prompt": selectors}
                model_specific_kwargs = {models[i]: {}}
                experiment = EXPERIMENTS[model_types[i]]([models[i]], input_kwargs, model_specific_kwargs)
            else:
                experiment = EXPERIMENTS[model_types[i]]([models[i]], selectors)
        else:
            if model_types[i] == "Replicate":
                input_kwargs = {"prompt": prompts}
                model_specific_kwargs = {models[i]: {}}
                experiment = EXPERIMENTS[model_types[i]]([models[i]], input_kwargs, model_specific_kwargs)
            else:
                experiment = EXPERIMENTS[model_types[i]]([models[i]], prompts)
        dfs.append(experiment.to_pandas_df(True, True))
    return dfs



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


try:
    import scipy.stats as stats
except ImportError:
    stats = None
import pandas


def ranking_correlation(
    row: pandas.core.series.Series, expected_ranking: list, ranking_column_name: str = "top doc ids"
) -> float:
    r"""
    A simple test that compares the expected ranking for a given query with the actual ranking produced
    by the embedding function being tested.

    Args:
        row (pandas.core.series.Series): A row of data from the full DataFrame (including input, model response, other
            metrics, etc).
        expected_ranking (list): the expected list of ranking to compare
        ranking_column_name (str): the column name of the actual ranking produced by the model,
            defaults to ``"top doc ids"``

    Example:
        >>> EXPECTED_RANKING_LIST = [
        >>>     ["id1", "id3", "id2"],
        >>>     ["id2", "id3", "id1"],
        >>>     ["id1", "id3", "id2"],
        >>>     ["id2", "id3", "id1"],
        >>> ]
        >>> experiment.evaluate("ranking_correlation", ranking_correlation, expected_ranking=EXPECTED_RANKING_LIST)
    """
    if stats is None:
        raise ModuleNotFoundError(
            "Package `SciPy` is required to be installed to use this evaluation method."
            "Please use `pip install scipy` to install the package"
        )
    actual_ranking = row[ranking_column_name]
    if len(expected_ranking) == 1 and len(actual_ranking) == 1:
        return 1.0 if expected_ranking == actual_ranking else -1.0
    correlation, _ = stats.spearmanr(actual_ranking, expected_ranking)
    return correlation



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.
import logging
import os

try:
    import anthropic
    from anthropic._types import NOT_GIVEN
except ImportError:
    anthropic = None
    NOT_GIVEN = None

from prompttools.selector.prompt_selector import PromptSelector
from prompttools.mock.mock import mock_anthropic_completion_fn
from .experiment import Experiment


class AnthropicCompletionExperiment(Experiment):
    r"""
    This class defines an experiment for Anthropic's completion API. It accepts lists for each argument
    passed into Anthropic's API, then creates a cartesian product of those arguments, and gets results for each.

    Note:
        - All arguments here should be a ``list``, even if you want to keep the argument frozen
          (i.e. ``temperature=[1.0]``), because the experiment will try all possible combination
          of the input arguments.
        - You should set ``os.environ["ANTHROPIC_API_KEY"] = YOUR_KEY`` in order to connect with Anthropic's API.

    Args:
        max_tokens_to_sample (list[int]):
            A list of integers representing The maximum number of tokens to generate before stopping.

        model (list[str]):
            the model(s) that will complete your prompt (e.g. "claude-2", "claude-instant-1")

        prompt (list[str]):
            Input prompt. For proper response generation you will need to format your prompt as follows:
            ``f"{HUMAN_PROMPT} USER_QUESTION {AI_PROMPT}"``, you can get built-in string by importing
            ``from anthropic HUMAN_PROMPT, AI_PROMPT``

        metadata (list):
            list of object(s) describing metadata about the request.

        stop_sequences (list[list[str]], optional):
            Sequences that will cause the model to stop generating completion text

        stream (list[bool], optional):
            Whether to incrementally stream the response using server-sent events.

        temperature (list[float], optional):
             The amount of randomness injected into the response

        top_k (list[int], optional):
            Only sample from the top K options for each subsequent token.

        top_p (list[float], optional):
            use nucleus sampling.

        timeout (list[float], optional):
           Override the client-level default timeout for this request, in seconds. Defaults to [600.0].
    """

    def __init__(
        self,
        model: list[str],
        prompt: list[str],
        metadata: list = [NOT_GIVEN],
        max_tokens_to_sample: list[int] = [1000],
        stop_sequences: list[list[str]] = [NOT_GIVEN],
        stream: list[bool] = [False],
        temperature: list[float] = [NOT_GIVEN],
        top_k: list[int] = [NOT_GIVEN],
        top_p: list[float] = [NOT_GIVEN],
        timeout: list[float] = [600.0],
        # These are extra parameters not available via kwargs
        # extra_headers=None,
        # extra_query=None,
        # extra_body=None,
    ):
        if anthropic is None:
            raise ModuleNotFoundError(
                "Package `anthropic` is required to be installed to use this experiment."
                "Please use `pip install anthropic` to install the package"
            )
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_anthropic_completion_fn
        else:
            self.client = anthropic.Anthropic(api_key=os.environ["ANTHROPIC_API_KEY"])
            self.completion_fn = self.anthropic_completion_fn

        # If we are using a prompt selector, we need to render
        # messages, as well as create prompt_keys to map the messages
        # to corresponding prompts in other models.
        if isinstance(prompt[0], PromptSelector):
            self.prompt_keys = {selector.for_anthropic(): selector.for_anthropic() for selector in prompt}
            prompt = [selector.for_anthropic() for selector in prompt]
        else:
            self.prompt_keys = prompt

        self.all_args = dict(
            max_tokens_to_sample=max_tokens_to_sample,
            model=model,
            prompt=prompt,
            metadata=metadata,
            stop_sequences=stop_sequences,
            stream=stream,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            timeout=timeout,
        )
        super().__init__()

    def anthropic_completion_fn(self, **input_args):
        try:
            return self.client.completions.create(**input_args)
        except anthropic.APIConnectionError as e:
            logging.error("The server could not be reached")
            logging.error(e.__cause__)  # an underlying Exception, likely raised within httpx.
        except anthropic.RateLimitError:
            logging.error("A 429 rate limit status code was received; we should back off a bit.")
        except anthropic.APIStatusError as e:
            logging.error("Another non-200-range status code was received")
            logging.error(e.status_code)
            logging.error(e.response)

    @staticmethod
    def _extract_responses(completion_response: "anthropic.types.Completion") -> list[str]:
        return [completion_response.completion]

    def _get_model_names(self):
        return [combo["model"] for combo in self.argument_combos]

    def _get_prompts(self):
        return [combo["prompt"] for combo in self.argument_combos]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


def chunk_text(text: str, max_chunk_length: int) -> list[str]:
    r"""
    Given a long string paragraph of text and a chunk max length, returns chunks of texts where each chunk's
    length is smaller than the max length, without breaking up individual words (separated by space).

    Args:
        text (str): source text to be chunked
        max_chunk_length (int): maximum length of a chunk
    """

    words = text.split()
    chunks = []
    current_chunk = ""

    for word in words:
        if len(current_chunk) + len(word) + 1 <= max_chunk_length:
            if current_chunk:
                current_chunk += " "
            current_chunk += word
        else:
            chunks.append(current_chunk)
            current_chunk = word

    if current_chunk:
        chunks.append(current_chunk)

    return chunks



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import os
from typing import Dict
import pandas.core.series
from .error import PromptToolsUtilityError

try:
    from pylint import epylint as lint
except ImportError:
    lint = None

PROMPTTOOLS_TMP = "prompttools_tmp.py"


def validate(text: str):
    r"""
    Validates that the generated text is python.

    Args:
        text (str): The generated text, which should be valid python.
    """
    if lint is None:
        raise RuntimeError(
            "Our built-in `validate_python` function requires pylint<3.0. Please use a custom eval function."
            "Feel free to open a GitHub issue or PR."
        )
    if os.path.isfile(PROMPTTOOLS_TMP):
        raise PromptToolsUtilityError
    with open(PROMPTTOOLS_TMP, "w") as f:
        f.write(text)
    pylint_stdout, _ = lint.py_run(PROMPTTOOLS_TMP, return_std=True)
    os.remove(PROMPTTOOLS_TMP)
    return 0.0 if "error" in pylint_stdout.getvalue() else 1.0


def validate_python_response(row: pandas.core.series.Series, response_column_name: str = "response") -> float:
    r"""
    Validate whether ``response`` string follows Python's syntax.

    Args:
        row (pandas.core.series.Series): A row of data from the full DataFrame (including input, model response, other
            metrics, etc).
        response_column_name (str): name of the column that contains the model's response, defaults to ``"response"``
    """
    return validate(row[response_column_name])


def evaluate(prompt: str, response: str, metadata: Dict) -> float:
    r"""
    Validate whether ``response`` string follows Python's syntax.

    Args:
        prompt (str): Not used.
        response (str): the string that will be validated
        metadata (dict): Not used.
    """
    return validate(response)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional, Type, Callable, Union
import jinja2
from .harness import ExperimentationHarness, Experiment
import copy


DOC_PROMPT_TEMPLATE = r"""Given these documents:{{documents}}

{{prompt}}
"""


def _doc_list_to_str(documents: list[str]) -> str:
    res = ""
    for d in documents:
        res += "\n"
        res += d
    return res


def _generate_doc_prompt(documents: list[str], prompt_or_msg: Union[str, list[dict[str, str]]], is_chat: bool):
    if not is_chat:
        prompt = prompt_or_msg
    else:  # You have a chat message object
        prompt = prompt_or_msg[-1]["content"]
    environment = jinja2.Environment()
    template = environment.from_string(DOC_PROMPT_TEMPLATE)
    doc_str = _doc_list_to_str(documents)

    doc_prompt = template.render(
        {
            "documents": doc_str,
            "prompt": prompt,
        }
    )
    if not is_chat:
        return doc_prompt
    else:
        new_msg = copy.deepcopy(prompt_or_msg)
        new_msg[-1]["content"] = doc_prompt
        return new_msg


class RetrievalAugmentedGenerationExperimentationHarness(ExperimentationHarness):
    r"""
    An experimentation harness used to test the Retrieval-Augmented Generation process, which
    involves a vector DB and a LLM at the same time.

    Args:
        vector_db_experiment (Experiment): An initialized vector DB experiment.
        llm_experiment_cls (Type[Experiment]): The experiment constructor that you would like to execute
            within the harness (e.g. ``prompttools.experiment.OpenAICompletionExperiment``)
        llm_arguments (dict[str, list]): Dictionary of arguments for the LLM.
        extract_document_fn (Callable): A function, when given a row of results from the vector DB experiment,
            extract the relevant documents (``list[str]``) that will be inserted into the template.
        extract_query_metadata_fn (Callable): A function, when given a row of results from the vector DB experiment,
            extract the relevant metadata and return a ``str`` that will be shown for visualization in the final
            result table
        prompt_template (str): A ``jinja``-styled templates, where documents and prompt will be inserted.
    """

    def __init__(
        self,
        vector_db_experiment: Experiment,
        llm_experiment_cls: Type[Experiment],
        llm_arguments: dict,
        extract_document_fn: Callable,
        extract_query_metadata_fn: Callable,
        prompt_template: str = DOC_PROMPT_TEMPLATE,
    ):
        self.vector_db_experiment = vector_db_experiment
        self.llm_experiment_cls: Type[Experiment] = llm_experiment_cls
        self.experiment: Optional[Experiment] = None
        self.llm_arguments = copy.copy(llm_arguments)
        self.extract_document_fn = extract_document_fn
        self.extract_query_metadata_fn = extract_query_metadata_fn
        self.prompt_templates = prompt_template

    def run(self) -> None:
        self.vector_db_experiment.run()
        document_lists: list[list[str]] = []
        # latencies = []  # TODO: Include latency results
        # Extract documents from the result of
        for i, row in self.vector_db_experiment.full_df.iterrows():
            document_lists.append(self.extract_document_fn(row))
            # latencies.append(row["latencies"])

        # Put documents into prompt template
        augmented_prompts = []
        is_chat = self.llm_experiment_cls._is_chat()
        input_arg_name = "messages" if is_chat else "prompt"
        for doc in document_lists:
            for prompt_or_msg in self.llm_arguments[input_arg_name]:
                augmented_prompts.append(_generate_doc_prompt(doc, prompt_or_msg, is_chat))

        # Pass documents into LLM
        self.llm_arguments[input_arg_name]: list[str] = augmented_prompts
        self.experiment = self.llm_experiment_cls(**self.llm_arguments)

        # Run the LLM experiment
        self.experiment.run()

        # Add "query text" (i.e. the prompt used to retrieve documents from the vector DB)
        # to the final results table here
        retrieval_n_rows = len(self.vector_db_experiment.full_df)
        query_metadata = [
            self.extract_query_metadata_fn(row) for _, row in self.vector_db_experiment.full_df.iterrows()
        ]
        final_n_row = len(self.full_df)

        self.partial_df["retrieval_metadata"] = [query_metadata[i % retrieval_n_rows] for i in range(final_n_row)]
        self.full_df["retrieval_metadata"] = self.partial_df["retrieval_metadata"]

    def visualize(self) -> None:
        if self.experiment is None:
            self.run()
        self.experiment.visualize()



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional, Type
from .harness import ExperimentationHarness, Experiment
import pandas as pd
from .utility import is_interactive
from IPython import display
from tabulate import tabulate
import logging


class SystemPromptExperimentationHarness(ExperimentationHarness):
    r"""
    An experimentation harness used to test various system prompts.

    Args:
        experiment (Type[Experiment]): The experiment that you would like to execute (e.g.
            ``prompttools.experiment.OpenAICompletionExperiment``)
        model_name (str): The name of the model.
        system_prompts (List[str]): A list of system prompts for the model
        human_messages (List[str]): A list of human (user) messages to pass into the model
        model_arguments (Optional[Dict[str, object]], optional): Additional arguments for the model.
            Defaults to ``None``. Note that the values are not lists.
    """

    _experiment_type = "Instruction"
    PIVOT_COLUMNS = ["system_prompt", "user_input"]

    def __init__(
        self,
        experiment: Type[Experiment],
        model_name: str,
        system_prompts: List[str],
        human_messages: List[str],
        model_arguments: Optional[Dict[str, object]] = None,
    ):
        self.experiment_cls_constructor = experiment
        self.model_name = model_name
        self.system_prompts = system_prompts
        self.human_messages = human_messages
        self.model_arguments = {} if model_arguments is None else model_arguments
        super().__init__()

    @staticmethod
    def _create_system_prompt(content: str) -> Dict[str, str]:
        return {"role": "system", "content": content}

    @staticmethod
    def _create_human_message(content: str) -> Dict[str, str]:
        return {"role": "user", "content": content}

    def prepare(self) -> None:
        r"""
        Creates messages to use for the experiment, and then initializes and prepares the experiment.
        """
        self.input_pairs_dict = {}
        messages_to_try = []
        for system_prompt in self.system_prompts:
            for message in self.human_messages:
                history = [
                    self._create_system_prompt(system_prompt),
                    self._create_human_message(message),
                ]
                messages_to_try.append(history)
                self.input_pairs_dict[str(history)] = (system_prompt, message)
        self.experiment = self.experiment_cls_constructor(
            [self.model_name],
            messages_to_try,
            **self._prepare_arguments(self.model_arguments),
        )
        super().prepare()

    def run(self, clear_previous_results: bool = False):
        if not self.experiment:
            self.prepare()
        super().run(clear_previous_results=clear_previous_results)

    def _get_state(self):
        state_params = {
            "experiment_cls_constructor": self.experiment_cls_constructor,
            "model_name": self.model_name,
            "system_prompts": self.system_prompts,
            "human_messages": self.human_messages,
            "model_arguments": self.model_arguments,
            "child_experiment_state": self.experiment._get_state() if self.experiment else None,
        }
        state = (
            state_params,
            self.experiment.full_df,
        )
        print("Creating state of experiment...")
        return state

    @classmethod
    def _load_state(cls, state, experiment_id: str, revision_id: str, experiment_type_str: str):
        (
            state_params,
            full_df,
        ) = state
        if experiment_type_str != cls._experiment_type:
            raise RuntimeError(
                f"The Experiment Type you are trying to load is {experiment_type_str},"
                "which does not match the current class."
            )

        experiment_cls_constructor = state_params["experiment_cls_constructor"]
        model_name = state_params["model_name"]
        system_prompts = state_params["system_prompts"]
        human_messages = state_params["human_messages"]
        model_arguments = state_params["model_arguments"]
        child_experiment_state = state_params["child_experiment_state"]

        harness = cls(experiment_cls_constructor, model_name, system_prompts, human_messages, model_arguments)
        harness.experiment = experiment_cls_constructor._load_state(
            child_experiment_state, None, None, experiment_cls_constructor._experiment_type
        )
        harness._experiment_id = experiment_id
        harness._revision_id = revision_id
        print("Loaded harness.")
        return harness

    def get_table(self, get_all_cols: bool = False) -> pd.DataFrame:
        columns_to_hide = [
            "stream",
            "response_id",
            "response_choices",
            "response_created",
            "response_created",
            "response_object",
            "response_model",
            "response_system_fingerprint",
            "revision_id",
            "log_id",
        ]

        if get_all_cols:
            return self.full_df
        else:
            table = self.full_df
            columns_to_hide.extend(
                [
                    col
                    for col in ["temperature", "top_p", "n", "presence_penalty", "frequency_penalty"]
                    if col not in self.partial_df.columns
                ]
            )
            for col in columns_to_hide:
                if col in table.columns:
                    table = table.drop(col, axis=1)
            return table

    def visualize(self, get_all_cols: bool = False):
        table = self.get_table(get_all_cols)
        if is_interactive():
            display.display(table)
        else:
            logging.getLogger().setLevel(logging.INFO)
            logging.info(tabulate(table, headers="keys", tablefmt="psql"))
